from zrb.helper.typing import (
    Any, Callable, Iterable, JinjaTemplate, List, Mapping, Optional, Union,
    TypeVar
)
from abc import ABC, abstractmethod
from zrb.task_env.env_file import EnvFile
from zrb.task_env.env import Env
from zrb.task_input.any_input import AnyInput

# flake8: noqa E501
TAnyTask = TypeVar('TAnyTask', bound='AnyTask')


class AnyTask(ABC):
    '''
    Abstract base class for defining tasks in a task management system.

    This class acts as a template for creating new task types. To define a new task, 
    extend this class and implement all its abstract methods. The `AnyTask` class is 
    considered atomic and is not broken into multiple interfaces.

    Subclasses should implement the abstract methods to define custom behavior for 
    task execution, state transitions, and other functionalities.
    '''

    @abstractmethod
    def copy(self) -> TAnyTask:
        '''
        Creates and returns a copy of the current task.

        The copied task can be modified using various setter methods like `set_name`, 
        `set_description`, and others, depending on the subclass implementation.

        Returns:
            TAnyTask: A copy of the current task.

        Examples:
            >>> from zrb import Task
            >>> task = Task(name='my-task', cmd='echo hello')
            >>> copied_task = task.copy()
            >>> copied_task.set_name('new_name')
        '''
        pass

    @abstractmethod
    async def run(self, *args: Any, **kwargs: Any) -> Any:
        '''
        Executes the main logic of the task.

        This asynchronous method should be implemented in subclasses to define the 
        task's primary functionality. The specific behavior and the return value 
        depend on the task's nature and purpose.

        Args:
            args (Any): Variable length argument list.
            kwargs (Any): Arbitrary keyword arguments.

        Returns:
            Any: The result of the task execution, the type of which is determined by 
            the specific task implementation.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def run(self, *args: Any, **kwargs: Any) -> int:
            >>>         self.print_out('Doing some calculation')
            >>>         return 42
        '''
        pass

    @abstractmethod
    async def check(self) -> bool:
        '''
        Checks if the current task is `ready`.

        Any other tasks depends on the current task, will be `started` once the current task is `ready`.

        This method should be implemented to define the criteria for considering the task 
        `ready`. The specifics of this completion depend on the task's 
        nature and the subclass implementation.

        Returns:
            bool: True if the task is completed, False otherwise.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def run(self, *args: Any, **kwargs: Any) -> int:
            >>>         self.print_out('Doing some calculation')
            >>>         self._completed = True
            >>>         return 42
            >>>     async def check(self) -> bool:
            >>>         return self._completed
        '''
        pass

    @abstractmethod
    def set_xcom(self, key: str, value: Any) -> str:
        '''
        Set xcom for cross task communication.

        Argss:
            key (str): Xcom key
            value (str): The value of the xcom

        Returns:
            str: Empty string
        
        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def run(self, *args: Any, **kwargs: Any) -> int:
            >>>         self.set_xcom('magic_word', 'hello')
            >>>         magic_word = self.get_xcom('magic_word')
            >>>         return 42
        '''
        pass

    @abstractmethod
    def set_task_xcom(self, key: str, value: Any) -> str:
        '''
        Set task xcom for cross task communication.

        Argss:
            key (str): Xcom key
            value (str): The value of the xcom

        Returns:
            str: Empty string
        
        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def run(self, *args: Any, **kwargs: Any) -> int:
            >>>         self.set_task_xcom('magic_word', 'hello')
            >>>         magic_word = self.get_xcom(f'{self.get_name()}.magic_word')
            >>>         return 42
        '''
        pass

    @abstractmethod
    def get_xcom(self, key: str) -> str:
        '''
        Get xcom value for cross task communication.

        Argss:
            key (str): Xcom key

        Returns:
            str: Value of xcom
        
        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def run(self, *args: Any, **kwargs: Any) -> int:
            >>>         return self.get_xcom('magic_word')
        '''
        pass

    @abstractmethod
    async def on_triggered(self):
        '''
        Defines actions to perform when the task status is set to `triggered`.

        Implement this method to specify behavior when the task transitions to the 
        `triggered` state. This could involve setting up prerequisites or sending 
        notifications.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def on_triggered(self):
            >>>         self.log_info('Task has been triggered')
        '''
        pass

    @abstractmethod
    async def on_waiting(self):
        '''
        Defines actions to perform when the task status is set to `waiting`.

        Implement this method to specify behavior when the task transitions to the 
        `waiting` state. This state usually indicates the task is waiting for some 
        condition or prerequisite to be met.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def on_waiting(self):
            >>>         self.log_info('Task is waiting to be started')
        '''
        pass

    @abstractmethod
    async def on_skipped(self):
        '''
        Defines actions to perform when the task status is set to `skipped`.

        Implement this method to specify behavior when the task is skipped. This could 
        include logging information, cleaning up resources, or any other necessary steps.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def on_skipped(self):
            >>>         self.log_info('Task was skipped') 
        '''
        pass

    @abstractmethod
    async def on_started(self):
        '''
        Defines actions to perform when the task status is set to 'started'.

        Implement this method to specify behavior when the task starts its execution. This 
        could involve initializing resources, logging, or other startup procedures.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def on_started(self):
            >>>         self.log_info('Task has started')
        '''
        pass

    @abstractmethod
    async def on_ready(self):
        '''
        Defines actions to be performed when the task status is `ready`.

        This asynchronous method should be implemented in subclasses to specify 
        actions that occur when the task reaches the `ready` state. This can include 
        any cleanup, notification, or follow-up actions specific to the task.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def on_ready(self):
            >>>         self.print_out('The task is ready')
        '''
        pass

    @abstractmethod
    async def on_failed(self, is_last_attempt: bool, exception: Exception):
        '''
        Specifies the behavior when the task execution fails.

        This asynchronous method should be implemented in subclasses to handle task 
        failure scenarios. It can include logging the error, performing retries, or 
        any other failure handling mechanisms.

        Args:
            is_last_attempt (bool): Indicates if this is the final retry attempt.
            exception (Exception): The exception that caused the task to fail.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def on_failed(self, is_last_attempt: bool, exception: Exception):
            >>>         if is_last_attempt:
            >>>             self.print_out('The task has failed with no remaining retries')
            >>>         else:
            >>>             self.print_out('The task failed, retrying...')
        '''
        pass

    @abstractmethod
    async def on_retry(self):
        '''
        Defines actions to perform when the task is retried.

        Implement this method to specify behavior when the task is retried after a failure. 
        This could include resetting states, logging the retry attempt, or other necessary 
        steps before re-execution.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def on_retry(self):
            >>>         self.log_info('Retrying task')
        '''
        pass

    @abstractmethod
    def to_function(
        self,
        env_prefix: str = '',
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True
    ) -> Callable[..., Any]:
        '''
        Converts the current task into a callable function.

        This method should be implemented to allow the task to be executed as a function. 
        Parameters can be used to modify the behavior of the generated function, such as 
        raising errors, asynchronous execution, and logging.

        Args:
            env_prefix (str): A prefix for environment variables.
            raise_error (bool): Whether to raise an error if the task execution fails.
            is_async (bool): Whether the resulting function should be asynchronous.
            show_done_info (bool): Whether to show information upon task completion.

        Returns:
            Callable[..., Any]: A callable representation of the task.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     async def run(self, *args: Any, **kwargs: Any) -> int:
            >>>         self.print_out('Doing some calculation')
            >>>         return 42
            >>> task = MyTask()
            >>> fn = task.to_function()
            >>> fn()
        '''
        pass

    @abstractmethod
    def insert_checker(self, *checkers: TAnyTask):
        '''
        Inserts one or more `AnyTask` instances at the beginning of the current task's checker list.

        This method is used to define dependencies for the current task. Tasks in the checker list are 
        executed before the current task. Adding a task to the beginning of the list means it will be 
        executed earlier than those already in the list.

        Args:
            checkers (TAnyTask): One or more task instances to be added to the checker list.

        Examples:
            >>> from zrb import Task
            >>> task = Task(name='task')
            >>> checker_task = Task(name='checker-task')
            >>> task.insert_checker(checker_task)
        '''
        pass

    @abstractmethod
    def add_checker(self, *checkers: TAnyTask):
        '''
        Adds one or more `AnyTask` instances to the end of the current task's checker list.

        This method appends tasks to the checker list, indicating that these tasks should be executed 
        before the current task, but after any tasks already in the checker list.

        Args:
            checkers (TAnyTask): One or more task instances to be added to the checker list.

        Examples:
            >>> from zrb import Task
            >>> task = Task(name='task')
            >>> checker_task = Task(name='checker-task')
            >>> task.add_checker(checker_task)
        '''
        pass

    @abstractmethod
    def insert_upstream(self, *upstreams: TAnyTask):
        '''
        Inserts one or more `AnyTask` instances at the beginning of the current task's upstream list.

        This method is used to define dependencies for the current task. Tasks in the upstream list are 
        executed before the current task. Adding a task to the beginning of the list means it will be 
        executed earlier than those already in the list.

        Args:
            upstreams (TAnyTask): One or more task instances to be added to the upstream list.

        Examples:
            >>> from zrb import Task
            >>> task = Task(name='task')
            >>> upstream_task = Task(name='upstream-task')
            >>> task.insert_upstream(upstream_task)
        '''
        pass

    @abstractmethod
    def add_upstream(self, *upstreams: TAnyTask):
        '''
        Adds one or more `AnyTask` instances to the end of the current task's upstream list.

        This method appends tasks to the upstream list, indicating that these tasks should be executed 
        before the current task, but after any tasks already in the upstream list.

        Args:
            upstreams (TAnyTask): One or more task instances to be added to the upstream list.

        Examples:
            >>> from zrb import Task
            >>> task = Task(name='task')
            >>> upstream_task = Task(name='upstream-task')
            >>> task.add_upstream(upstream_task)
        '''
        pass

    @abstractmethod
    def insert_input(self, *inputs: AnyInput):
        '''
        Inserts one or more `AnyInput` instances at the beginning of the current task's input list.

        This method is used to add inputs that the task will process. Inserting an input at the beginning 
        of the list gives it precedence over those already present.

        Args:
            inputs (AnyInput): One or more input instances to be added to the input list.

        Examples:
            >>> from zrb import Task, Input
            >>> task = Task(name='task')
            >>> email_input = Input(name='email-address')
            >>> task.insert_input(email_input)
        '''
        pass

    @abstractmethod
    def add_input(self, *inputs: AnyInput):
        '''
        Adds one or more `AnyInput` instances to the end of the current task's input list.

        This method is used to append inputs for the task to process, placing them after any inputs 
        already specified.

        Args:
            inputs (AnyInput): One or more input instances to be added to the input list.

        Examples:
            >>> from zrb import Task, Input
            >>> task = Task(name='task')
            >>> email_input = Input(name='email-address')
            >>> task.add_input(email_input)
        '''
        pass

    @abstractmethod
    def insert_env(self, *envs: Env):
        '''
        Inserts one or more `Env` instances at the beginning of the current task's environment variable list.

        This method allows for setting or overriding environment variables for the task, with earlier entries 
        having precedence over later ones.

        Args:
            envs (Env): One or more environment variable instances to be added.

        Examples:
            >>> from zrb import Task, Env
            >>> task = Task(name='task')
            >>> db_url_env = Env(name='DATABASE_URL', value='postgresql://...')
            >>> task.insert_env(env_var)
        '''
        pass

    @abstractmethod
    def add_env(self, *envs: Env):
        '''
        Adds one or more `Env` instances to the end of the current task's environment variable list.

        Use this method to append environment variables for the task, which will be used after 
        any variables already set.

        Args:
            envs (Env): One or more environment variable instances to be added.
        
        Examples:
            >>> from zrb import Task, Env
            >>> task = Task(name='task')
            >>> db_url_env = Env(name='DATABASE_URL', value='postgresql://...')
            >>> task.add_env(env_var)
        '''
        pass

    @abstractmethod
    def insert_env_file(self, *env_files: EnvFile):
        '''
        Inserts one or more `EnvFile` instances at the beginning of the current task's environment file list.

        This method is used to specify environment variable files whose contents should be loaded 
        before those of any files already in the list. This is useful for overriding or setting 
        additional environment configurations.

        Args:
            env_files (EnvFile): One or more environment file instances to be added.

        Examples:
            >>> from zrb import Task, EnvFile
            >>> task = Task()
            >>> env_file = EnvFile(path='config.env')
            >>> task.insert_env_file(env_file)
        '''
        pass

    @abstractmethod
    def add_env_file(self, *env_files: EnvFile):
        '''
        Adds one or more `EnvFile` instances to the end of the current task's environment file list.

        Use this method to append environment file references, which will be processed after 
        any files already specified. This allows for supplementing the existing environment 
        configuration.

        Args:
            env_files (EnvFile): One or more environment file instances to be added.

        Examples:
            >>> from zrb import Task, EnvFile
            >>> task = Task()
            >>> env_file = EnvFile(path='config.env')
            >>> task.add_env_file(env_file)
        '''
        pass

    @abstractmethod
    def _set_execution_id(self, execution_id: str):
        '''
        Sets the execution ID for the current task.

        This method is intended for internal use to assign a unique identifier to the task's execution. 
        This ID can be used for tracking, logging, and inter-task communication.

        This method should not be used externally, as it is meant to be managed within the task system.

        Args:
            execution_id (str): A string representing the unique execution ID.
        '''
        pass

    @abstractmethod
    def set_name(self, new_name: str):
        '''
        Sets a new name for the current task.

        This method is used to update the task's name, typically after creating a copy of an existing task. 
        The new name helps in differentiating the task in the task management system.

        Args:
            new_name (str): A string representing the new name to be assigned to the task.
        '''
        pass

    @abstractmethod
    def set_description(self, new_description: str):
        '''
        Sets a new description for the current task.

        This method allows updating the task's description to provide more context or details about its purpose and behavior. 
        Useful for enhancing clarity and maintainability in the task management system.

        Args:
            new_description (str): A string representing the new description of the task.
        '''
        pass

    @abstractmethod
    def set_icon(self, new_icon: str):
        '''
        Assigns a new icon to the current task.

        This method is used for setting or updating the task's icon, which can be utilized for visual representation 
        in a user interface. The icon should ideally be a string identifier that maps to an actual graphical resource.

        Args:
            new_icon (str): A string representing the icon identifier for the task.
        '''
        pass

    @abstractmethod
    def set_color(self, new_color: str):
        '''
        Defines a new color for the current task.

        This method updates the color associated with the task. This can be useful for categorization, 
        priority indication, or visual differentiation in a UI.

        Args:
            new_color (str): A string representing the color to be assigned to the task. 
        '''
        pass

    @abstractmethod
    def set_should_execute(
        self, should_execute: Union[bool, str, Callable[..., bool]]
    ):
        '''
        Determines whether the task should execute.

        This method configures the execution criteria for the task. It can be set as a boolean value, 
        a string representing a condition, or a callable that returns a boolean. This is useful for 
        conditional task execution based on dynamic criteria.

        Args:
            should_execute (Union[bool, str, Callable[..., bool]]): The condition to determine if the task should execute. 
        '''
        pass

    @abstractmethod
    def set_retry(self, new_retry: int):
        '''
        Sets the number of retry attempts for the task.

        This method configures how many times the task should be retried in case of failure. 
        It's essential for tasks that may fail transiently and need multiple attempts for successful execution.

        Args:
            new_retry (int): An integer representing the number of retry attempts.
        '''
        pass

    @abstractmethod
    def set_retry_interval(self, new_retry_interval: Union[float, int]):
        '''
        Specifies the interval between retry attempts for the task.

        This method sets the duration to wait before retrying the task after a failure. 
        This can help in scenarios where immediate retry is not desirable or effective.

        Args:
            new_retry_interval (Union[float, int]): The time interval (in seconds) to wait before a retry attempt. 
        '''
        pass

    @abstractmethod
    def set_checking_interval(self, new_checking_interval: Union[float, int]):
        '''
        Sets the interval for checking the task's readiness or completion status.

        This method defines how frequently the system should check if the task is ready or completed. 
        It's useful for tasks that have an indeterminate completion time.

        Args:
            new_checking_interval (Union[float, int]): The time interval (in seconds) for readiness or checks. 
        '''
        pass

    @abstractmethod
    def get_execution_id(self) -> str:
        '''
        Retrieves the execution ID of the task.

        This method returns the unique identifier associated with the task's execution. 
        The execution ID is crucial for tracking, logging, and differentiating between 
        multiple instances or runs of the same task.

        Returns:
            str: The unique execution ID of the task.
        '''
        pass

    @abstractmethod
    def get_icon(self) -> str:
        '''
        Retrieves the icon identifier of the current task.

        This method is used to get the icon associated with the task, which can be utilized for 
        visual representation in user interfaces or documentation.

        Returns:
            str: A string representing the icon identifier for the task
        '''
        pass

    @abstractmethod
    def get_color(self) -> str:
        '''
        Retrieves the color associated with the current task.

        This method returns the color of the task, useful for visual differentiation, priority indication, 
        or categorization in user interfaces or documentation.

        Returns:
            str: A string representing the color assigned to the task.
        '''
        pass

    @abstractmethod
    def get_description(self) -> str:
        '''
        Fetches the current description of the task.

        This method is used to obtain the detailed description of the task, providing insights into its purpose, 
        functionality, and usage within the task management system.

        Returns:
            str: The description of the task.
        '''
        pass

    @abstractmethod
    def get_name(self) -> str:
        '''
        Get task name

        Returns:
            str: name of the task
        '''
        pass

    @abstractmethod
    def get_cli_name(self) -> str:
        '''
        Gets the command-line interface (CLI) name of the task.

        This method returns the name used to invoke the task via a CLI, facilitating integration with command-line tools 
        or scripts.

        Returns:
            str: The CLI name of the task.
        '''
        pass

    @abstractmethod
    def _get_full_cli_name(self) -> str:
        '''
        Retrieves the full command-line interface (CLI) name of the task.

        Intended for internal use, this method provides the complete CLI name, including any 
        prefixes or namespaces, used primarily for logging or debugging purposes.

        Returns:
            str: The full CLI name of the task.
        '''
        pass

    @abstractmethod
    def _set_has_cli_interface(self):
        '''
        Marks the task as having a CLI interface.

        This internal method is used to indicate that the task is accessible and executable through a CLI, 
        enabling the task system to appropriately handle its CLI interactions.
        '''
        pass

    @abstractmethod
    def inject_env_files(self):
        '''
        Injects additional `EnvFile` into the task.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     def inject_env_files(self):
            >>>         self.add_env_files(EnvFile(path='config.env'))
        '''
        pass

    @abstractmethod
    def _get_env_files(self) -> List[EnvFile]:
        '''
        Retrieves the list of environment variable files associated with the task.

        Intended for internal use, this method returns a list of `EnvFile` instances that the task 
        uses to load environment variables, primarily for setup and configuration purposes.

        Returns:
            List[EnvFile]: A list of `EnvFile` instances associated with the task.
        '''
        pass

    @abstractmethod
    def inject_envs(self):
        '''
        Injects environment variables into the task.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     def inject_envs(self):
            >>>         self.add_envs(Env(name='DATABASE_URL'))
        '''
        pass

    @abstractmethod
    def _get_envs(self) -> List[Env]:
        '''
        Retrieves the list of environment variables set for the task.

        For internal use, this method returns a list of `Env` instances representing the environment variables 
        configured for the task, essential for understanding and debugging the task's environment setup.

        Returns:
            List[Env]: A list of `Env` instances representing the environment variables of the task.
        '''
        pass

    @abstractmethod
    def inject_inputs(self):
        '''
        Injects custom inputs into the task.

        This method is used to programmatically add input parameters to the task, allowing 
        dynamic customization of the task's input data. Subclasses should override this method 
        to define specific inputs that the task should receive.

        Examples:
            >>> from zrb import Task, Input
            >>> class MyTask(Task):
            >>>     def inject_inputs(self):
            >>>         self.add_input(Input(name='user_email', type='email'))
        '''
        pass

    @abstractmethod
    def _get_inputs(self) -> List[AnyInput]:
        '''
        Retrieves the list of inputs associated with the task.

        This internal method is used to obtain all the inputs that have been set for the task, 
        either through static definition or via the `inject_inputs` method. It's primarily used 
        for introspection and debugging purposes.

        Returns:
            List[AnyInput]: A list of `AnyInput` instances representing the inputs for the task.
        '''
        pass

    @abstractmethod
    def inject_checkers(self):
        '''
        Injects custom checkers into the task.

        This method allows for the addition of custom validation or condition checkers. These 
        checkers can be used to verify certain conditions before the task execution proceeds. 
        Subclasses should implement this method to define task-specific checkers.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     def inject_checkers(self):
            >>>         self.add_checker(some_custom_condition_checker)
        '''
        pass

    @abstractmethod
    def _get_checkers(self) -> Iterable[TAnyTask]:
        '''
        Retrieves the checkers set for the task.

        This internal method returns an iterable of all the checkers that have been added to 
        the task. It's mainly used for internal logic and debugging to understand the 
        validations or conditions applied to the task.

        Returns:
            Iterable[TAnyTask]: An iterable of checkers associated with the task.
        '''
        pass

    @abstractmethod
    def inject_upstreams(self):
        '''
        Injects upstream tasks into the current task.

        This method is used for programmatically adding upstream dependencies to the task. 
        Upstream tasks are those that must be completed before the current task starts. 
        Override this method in subclasses to specify such dependencies.

        Examples:
            >>> from zrb import Task
            >>> class MyTask(Task):
            >>>     def inject_upstreams(self):
            >>>         self.add_upstream(another_task)
        '''
        pass

    @abstractmethod
    def _get_upstreams(self) -> Iterable[TAnyTask]:
        '''
        Retrieves the upstream tasks of the current task.

        An internal method to get the list of upstream tasks that have been set for the 
        task, either statically or through `inject_upstreams`. This is essential for task 
        scheduling and dependency management.

        Returns:
            Iterable[TAnyTask]: An iterable of upstream tasks.
        '''
        pass

    @abstractmethod
    def _get_combined_inputs(self) -> Iterable[AnyInput]:
        '''
        Combines and retrieves all inputs for the task.

        This internal method aggregates inputs from various sources (static definition, 
        `inject_inputs`, etc.) and provides a unified view of all inputs that the task 
        will process. This is crucial for preparing the task's runtime environment.

        Returns:
            Iterable[AnyInput]: An iterable of all combined inputs for the task.
        '''
        pass

    @abstractmethod
    def log_debug(self, message: Any):
        '''
        Log message with log level "DEBUG"

        You can set Zrb log level by using `ZRB_LOGGING_LEVEL` environment
        '''
        pass

    @abstractmethod
    def log_warn(self, message: Any):
        '''
        Log message with log level "WARNING"

        You can set Zrb log level by using `ZRB_LOGGING_LEVEL` environment
        '''
        pass

    @abstractmethod
    def log_info(self, message: Any):
        '''
        Log message with log level "INFO"

        You can set Zrb log level by using `ZRB_LOGGING_LEVEL` environment
        '''
        pass

    @abstractmethod
    def log_error(self, message: Any):
        '''
        Log message with log level "ERROR"

        You can set Zrb log level by using `ZRB_LOGGING_LEVEL` environment
        '''
        pass

    @abstractmethod
    def log_critical(self, message: Any):
        '''
        Log message with log level "CRITICAL"

        You can set Zrb log level by using `ZRB_LOGGING_LEVEL` environment
        '''
        pass

    @abstractmethod
    def print_out(self, message: Any, trim_message: bool = True):
        '''
        Print message to stderr as normal text.
        '''
        pass

    @abstractmethod
    def print_err(self, message: Any, trim_message: bool = True):
        '''
        Print message to stderr and style it as error.
        '''
        pass

    @abstractmethod
    def print_out_dark(self, message: Any, trim_message: bool = True):
        '''
        Print message to stdout and style it as faint.
        '''
        pass

    @abstractmethod
    def get_input_map(self) -> Mapping[str, Any]:
        '''
        Get a map representing task's Inputs.

        Typically used inside `run`, `check`, or in `@python_task` decorator

        Examples:
            >>> from zrb import python_task, Task, Input
            >>> @python_task(name='task', inputs=[Input(name='name')])
            >>> def task(*args, **kwargs):
            >>>     task: Task = kwargs.get('_task')
            >>>     for key, value in task.get_input_map():
            >>>         task.print_out(f'{key}: {value}')
        '''
        pass

    @abstractmethod
    def get_env_map(self) -> Mapping[str, Any]:
        '''
        Get a map representing task's Envs and EnvFiles

        Typically used inside `run`, `check`, or in `@python_task` decorator

        Examples:
            >>> from zrb import python_task, Task, Env
            >>> @python_task(name='task', envs=[Env(name='DB_URL')])
            >>> def task(*args, **kwargs):
            >>>     task: Task = kwargs.get('_task')
            >>>     for key, value in task.get_env_map():
            >>>         task.print_out(f'{key}: {value}')
        '''
        pass

    @abstractmethod
    def render_any(
        self, value: Any, data: Optional[Mapping[str, Any]] = None
    ) -> Any:
        '''
        Render any value.
        '''
        pass

    @abstractmethod
    def render_float(
        self,
        value: Union[JinjaTemplate, float],
        data: Optional[Mapping[str, Any]] = None
    ) -> float:
        '''
        Render float value.
        '''
        pass

    @abstractmethod
    def render_int(
        self,
        value: Union[JinjaTemplate, int],
        data: Optional[Mapping[str, Any]] = None
    ) -> int:
        pass

    @abstractmethod
    def render_bool(
        self, 
        value: Union[JinjaTemplate, bool],
        data: Optional[Mapping[str, Any]] = None
    ) -> bool:
        '''
        Render int value.
        '''
        pass

    @abstractmethod
    def render_str(
        self,
        value: JinjaTemplate,
        data: Optional[Mapping[str, Any]] = None
    ) -> str:
        '''
        Render str value.
        '''
        pass

    @abstractmethod
    def render_file(
        self,
        path: JinjaTemplate,
        data: Optional[Mapping[str, Any]] = None
    ) -> str:
        '''
        Render file content.
        '''
        pass

    @abstractmethod
    def _run_all(self, *args: Any, **kwargs: Any) -> Any:
        '''
        For internal use.

        Run this task and all its upstreams.
        '''
        pass

    @abstractmethod
    def _loop_check(self, show_info: bool) -> bool:
        '''
        For internal use.
        
        Regularly check whether the task is ready or not.
        '''
        pass

    @abstractmethod
    def _set_keyval(self, kwargs: Mapping[str, Any], env_prefix: str):
        '''
        For internal use.

        Set current task's key values.
        '''
        pass

    @abstractmethod
    def _print_result(self, result: Any):
        '''
        For internal use.

        Call `print_result` or print values based on result type and other conditions.
        '''
        pass

    @abstractmethod
    def print_result(self, result: Any):
        '''
        Print the task result to stdout for further processing.

        Override this method in subclasses to customize how the task result is displayed 
        or processed. Useful for integrating the task output with other systems or 
        command-line tools.

        Args:
            result (Any): The result of the task to be printed.

        Examples:
            >>> from zrb import Task
            >>> # Example of overriding in a subclass
            >>> class MyTask(Task):
            >>>    def print_result(self, result: Any):
            >>>        print(f'Result: {result}')
        '''
        pass



from zrb.builtin.base64 import encode, decode


def test_encode():
    function = encode.to_function()
    result = function(text='Philosopher Stone')
    assert result == 'UGhpbG9zb3BoZXIgU3RvbmU='


def test_decode():
    function = decode.to_function()
    result = function(text='UGhpbG9zb3BoZXIgU3RvbmU=')
    assert result == 'Philosopher Stone'



from zrb.task.checker import Checker
from zrb.task.task import Task
from zrb.task_input.str_input import StrInput
from zrb.task_env.env import Env
import asyncio


class DelayedChecker(Checker):
    async def inspect(self, *args, **kwargs) -> bool:
        await asyncio.sleep(1)
        return True


def test_use_the_same_task_for_checker_and_upstream():
    delayed_checker = DelayedChecker()
    # use delayed_chacker as task_1 checker
    task_1 = Task(
        name='task-1',
        inputs=[StrInput(name='input-1')],
        envs=[Env(name='ENV1')],
        checkers=[delayed_checker]
    )
    # use delayed_chacker as task_2 checker
    task_2 = Task(
        name='task-2',
        inputs=[StrInput(name='input-2')],
        envs=[Env(name='ENV2')],
    )
    task_2.add_checker(delayed_checker)
    # use delayed_chacker as task_3 upstream
    task = Task(
        name='task',
        upstreams=[task_1, task_2, delayed_checker]
    )
    fn = task.to_function()
    fn()



from zrb import CmdTask, EnvFile, BoolInput, StrInput, IntInput, runner
from zrb.builtin.group import project_group
from ._config import (
    CURRENT_DIR, LOAD_TEST_DIR, LOAD_TEST_TEMPLATE_ENV_FILE_NAME
)
import os

###############################################################################
# ⚙️ prepare-kebab-zrb-task-name-load-test
###############################################################################

prepare_snake_zrb_app_name_load_test = CmdTask(
    icon='🚤',
    name='prepare-kebab-zrb-app-name-load-test',
    description='Prepare load test for human readable zrb app name',
    group=project_group,
    cwd=LOAD_TEST_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'activate-venv.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'app-prepare-load-test.sh'),
    ]
)
runner.register(prepare_snake_zrb_app_name_load_test)

###############################################################################
# ⚙️ load-test-kebab-zrb-task-name
###############################################################################

load_test_snake_zrb_app_name_load_test = CmdTask(
    icon='🧪',
    name='load-test-kebab-zrb-app-name',
    description='Load test human readable zrb app name',
    group=project_group,
    upstreams=[prepare_snake_zrb_app_name_load_test],
    inputs=[
        BoolInput(
            name='kebab-zrb-app-name-load-test-headless',
            default=True,
            description='Load test UI headless',
            prompt='Load test UI headless (if True, there will be no UI)',
        ),
        IntInput(
            name='kebab-zrb-app-name-load-test-port',
            default=8089,
            description='Load test UI web port',
            prompt='Load test UI web port (Only works if headless is False)',
        ),
        IntInput(
            name='kebab-zrb-app-name-load-test-users',
            default=200,
            description='Load test users',
            prompt='Load test users',
        ),
        IntInput(
            name='kebab-zrb-app-name-load-test-spawn-rate',
            default=10,
            description='Load test spawn rate',
            prompt='Load test spawn rate',
        ),
        StrInput(
            name='kebab-zrb-app-name-load-test-url',
            default='http://localhost:zrbAppHttpPort',
            description='Load test url',
            prompt='Load test url',
        )
    ],
    cwd=LOAD_TEST_DIR,
    env_files=[
        EnvFile(
            path=LOAD_TEST_TEMPLATE_ENV_FILE_NAME,
            prefix='LOAD_TEST_ZRB_ENV_PREFIX'
        )
    ],
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'activate-venv.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'app-load-test.sh'),
    ]
)
runner.register(load_test_snake_zrb_app_name_load_test)



'''A Kubernetes Python Pulumi program to deploy kebab-zrb-app-name'''

from pulumi_kubernetes.helm.v3 import Chart, LocalChartOpts
import pulumi
import os

CURRENT_DIR = os.path.dirname(__file__)
NAMESPACE = os.getenv('NAMESPACE', 'default')

image = os.getenv('IMAGE', 'kebab-zrb-app-name:latest')
replica = int(os.getenv('REPLICA', '1'))
app_labels = {'app': 'kebab-zrb-app-name'}

# https://www.pulumi.com/registry/packages/kubernetes/api-docs/helm/v3/chart/#local-chart-directory  # noqa
release = Chart(
    'kebab-zrb-app-name',
    LocalChartOpts(
        path='./helm-charts/kebab-zrb-app-name',
        namespace=NAMESPACE,
        values={
        }
    )
)

pulumi.export('helm-release', release.resources)



from zrb import CmdTask, Env, EnvFile, PathChecker, runner
from zrb.builtin.group import project_group
from ._config import (
    APP_FRONTEND_DIR, APP_FRONTEND_BUILD_DIR, CURRENT_DIR,
    APP_TEMPLATE_ENV_FILE_NAME
)
import os

###############################################################################
# ⚙️ build-kebab-zrb-task-name-frontend
###############################################################################

build_snake_zrb_app_name_frontend = CmdTask(
    icon='🚤',
    name='build-kebab-zrb-app-name-frontend',
    description='Build frontend for human readable zrb app name',
    group=project_group,
    cwd=APP_FRONTEND_DIR,
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'app-build-frontend.sh'),
    checkers=[
        PathChecker(
            name='check-kebab-zrb-app-name-frontend-build',
            path=APP_FRONTEND_BUILD_DIR
        )
    ],
    env_files=[
        EnvFile(
            path=APP_TEMPLATE_ENV_FILE_NAME, prefix='ZRB_ENV_PREFIX'
        ),
    ],
    envs=[
        Env(name='WATCH', os_name='', default='1')
    ]
)
runner.register(build_snake_zrb_app_name_frontend)

###############################################################################
# ⚙️ build-kebab-zrb-task-name-frontend-once
###############################################################################

build_snake_zrb_app_name_frontend_once = CmdTask(
    icon='🚤',
    name='build-kebab-zrb-app-name-frontend_once',
    description='Build frontend for human readable zrb app name',
    cwd=APP_FRONTEND_DIR,
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'app-build-frontend.sh'),
    checkers=[
        PathChecker(
            name='check-kebab-zrb-app-name-frontend-build',
            path=APP_FRONTEND_BUILD_DIR
        )
    ],
    env_files=[
        EnvFile(
            path=APP_TEMPLATE_ENV_FILE_NAME, prefix='ZRB_ENV_PREFIX'
        ),
    ],
    envs=[
        Env(name='WATCH', os_name='', default='0')
    ]
)



from typing import Mapping
from zrb import Env, ServiceConfig, EnvFile
from zrb.helper.util import to_kebab_case

import jsons
import os

CURRENT_DIR = os.path.dirname(__file__)
PROJECT_DIR = os.path.abspath(os.path.join(CURRENT_DIR, '..', '..'))
RESOURCE_DIR = os.path.join(PROJECT_DIR, 'src', 'kebab-zrb-app-name')
DEPLOYMENT_DIR = os.path.join(RESOURCE_DIR, 'deployment')
DEPLOYMENT_TEMPLATE_ENV_FILE_NAME = os.path.join(
    DEPLOYMENT_DIR, 'template.env'
)
APP_DIR = os.path.join(RESOURCE_DIR, 'src')
APP_FRONTEND_DIR = os.path.join(APP_DIR, 'frontend')
APP_FRONTEND_BUILD_DIR = os.path.join(APP_FRONTEND_DIR, 'build')
APP_TEMPLATE_ENV_FILE_NAME = os.path.join(APP_DIR, 'template.env')

LOAD_TEST_DIR = os.path.join(RESOURCE_DIR, 'loadtest')
LOAD_TEST_TEMPLATE_ENV_FILE_NAME = os.path.join(RESOURCE_DIR, 'template.env')

MODULE_CONFIG_PATH = os.path.join(CURRENT_DIR, 'config', 'modules.json')

with open(MODULE_CONFIG_PATH) as file:
    MODULE_JSON_STR = file.read()
MODULES = jsons.loads(MODULE_JSON_STR)

DOCKER_COMPOSE_APP_ENV_FILE_NAME = os.path.join(
    RESOURCE_DIR, 'docker-compose-app.env'
)

SERVICE_CONFIGS: Mapping[str, ServiceConfig] = {}
service_names = [to_kebab_case(module) + '-service' for module in MODULES]
for suffix in ['', 'gateway'] + service_names:
    service_suffix = '-' + suffix if suffix != '' else ''
    service_name = f'kebab-zrb-app-name{service_suffix}'
    SERVICE_CONFIGS[service_name] = ServiceConfig(
        env_files=[
            EnvFile(
                path=APP_TEMPLATE_ENV_FILE_NAME,
                prefix='CONTAINER_ZRB_ENV_PREFIX'
            ),
            EnvFile(
                path=DOCKER_COMPOSE_APP_ENV_FILE_NAME,
                prefix='CONTAINER_ZRB_ENV_PREFIX'
            )
        ],
        envs=[
            Env(
                name='APP_OTEL_EXPORTER_OTLP_ENDPOINT',
                os_name='',
                default='http://otel-collector:4317'
            )
        ]
    )



from zrb.helper.typing import (
    Any, Callable, Iterable, List, Mapping, Optional, Union
)
from zrb.helper.callable import run_async
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnTriggered, OnWaiting, OnSkipped, OnStarted, OnReady, OnRetry, OnFailed
)
from zrb.task.base_task.component.trackers import (
    AttemptTracker, FinishTracker
)
from zrb.task.base_task.component.renderer import Renderer
from zrb.task.base_task.component.base_task_model import BaseTaskModel
from zrb.task.parallel import AnyParallel
from zrb.advertisement import advertisements
from zrb.task_group.group import Group
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_input.any_input import AnyInput
from zrb.helper.accessories.name import get_random_name
from zrb.helper.advertisement import get_advertisement
from zrb.helper.string.conversion import to_variable_name
from zrb.helper.string.modification import double_quote
from zrb.helper.map.conversion import to_str as map_to_str
from zrb.config.config import show_advertisement

import asyncio
import copy


@typechecked
class BaseTask(
    FinishTracker, AttemptTracker, Renderer, BaseTaskModel, AnyTask
):
    '''
    Base class for all tasks.
    Every task definition should be extended from this class.
    '''
    __xcom: Mapping[str, Mapping[str, str]] = {}

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        description: str = '',
        inputs: List[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0,
        run: Optional[Callable[..., Any]] = None,
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
        return_upstream_result: bool = False
    ):
        # init properties
        retry_interval = retry_interval if retry_interval >= 0 else 0
        checking_interval = checking_interval if checking_interval > 0 else 0.1
        retry = retry if retry >= 0 else 0
        # init parent classes
        FinishTracker.__init__(self)
        Renderer.__init__(self)
        AttemptTracker.__init__(self, retry=retry)
        BaseTaskModel.__init__(
            self,
            name=name,
            group=group,
            description=description,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            retry=retry,
            retry_interval=retry_interval,
            upstreams=upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            run=run,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            should_execute=should_execute,
            return_upstream_result=return_upstream_result
        )
        # init private flags
        self.__is_keyval_set = False
        self.__is_check_triggered: bool = False
        self.__is_ready: bool = False
        self.__is_execution_triggered: bool = False
        self.__is_execution_started: bool = False

    def __rshift__(
        self, operand: Union[AnyParallel, AnyTask]
    ) -> Union[AnyParallel, AnyTask]:
        if isinstance(operand, AnyTask):
            operand.add_upstream(self)
            return operand
        if isinstance(operand, AnyParallel):
            other_tasks: List[AnyTask] = operand.get_tasks()
            for other_task in other_tasks:
                other_task.add_upstream(self)
            return operand

    def set_task_xcom(self, key: str, value: Any) -> str:
        return self.set_xcom(
            key='.'.join([self.get_name(), key]),
            value=value
        )

    def set_xcom(self, key: str, value: Any) -> str:
        execution_id = self.get_execution_id()
        if execution_id not in self.__xcom:
            self.__xcom[execution_id] = {}
        execution_id = self.get_execution_id()
        self.__xcom[execution_id][key] = f'{value}'
        return ''

    def get_xcom(self, key: str) -> str:
        execution_id = self.get_execution_id()
        if execution_id not in self.__xcom:
            return ''
        return self.__xcom[execution_id].get(key, '')

    def clear_xcom(self, execution_id: str = '') -> str:
        if execution_id == '':
            execution_id = self.get_execution_id()
        if execution_id in self.__xcom:
            del self.__xcom[execution_id]
            return ''

    def copy(self) -> AnyTask:
        return copy.deepcopy(self)

    def to_function(
        self,
        env_prefix: str = '',
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True
    ) -> Callable[..., Any]:
        async def function(*args: Any, **kwargs: Any) -> Any:
            self.log_info('Copy task')
            self_cp = self.copy()
            return await self_cp._run_and_check_all(
                env_prefix=env_prefix,
                raise_error=raise_error,
                args=args,
                kwargs=kwargs,
                show_done_info=show_done_info
            )
        if is_async:
            return function
        return lambda *args, **kwargs: asyncio.run(function(*args, **kwargs))

    async def run(self, *args: Any, **kwargs: Any) -> Any:
        if self._run_function is not None:
            return await run_async(self._run_function, *args, **kwargs)
        return None

    async def on_triggered(self):
        self.log_info('State: triggered')
        execution_id = self.get_execution_id()
        self.log_info(f'Execution id: {execution_id}')
        if self._on_triggered is not None:
            await run_async(self._on_triggered, self)

    async def on_waiting(self):
        self.log_info('State: waiting')
        if self._on_waiting is not None:
            await run_async(self._on_waiting, self)

    async def on_skipped(self):
        self.log_info('State: skipped')
        if self._on_skipped is not None:
            await run_async(self._on_skipped, self)

    async def on_started(self):
        self.log_info('State: started')
        if self._on_started is not None:
            await run_async(self._on_started, self)

    async def on_ready(self):
        self.log_info('State: ready')
        if self._on_ready is not None:
            await run_async(self._on_ready, self)

    async def on_failed(self, is_last_attempt: bool, exception: Exception):
        failed_state_message = 'State failed'
        if is_last_attempt:
            failed_state_message = 'State failed (last attempt)'
        self.log_info(failed_state_message)
        if self._on_failed is not None:
            await run_async(self._on_failed, self, is_last_attempt, exception)

    async def on_retry(self):
        self.log_info('State: retry')
        if self._on_retry is not None:
            await run_async(self._on_retry, self)

    async def check(self) -> bool:
        return await self._is_done()

    async def _run_and_check_all(
        self,
        env_prefix: str,
        raise_error: bool,
        args: Iterable[Any],
        kwargs: Mapping[str, Any],
        show_done_info: bool = True
    ):
        try:
            self._start_timer()
            if self.get_execution_id() == '':
                self._set_execution_id(
                    get_random_name(add_random_digit=True, digit_count=5)
                )
                self._propagate_execution_id()
            self.log_info('Set input and env map')
            await self._set_keyval(kwargs=kwargs, env_prefix=env_prefix)
            self.log_info('Set run kwargs')
            new_kwargs = self.get_input_map()
            # make sure args and kwargs['_args'] are the same
            self.log_info('Set run args')
            new_args = copy.deepcopy(args)
            if len(args) == 0 and '_args' in kwargs:
                new_args = kwargs['_args']
            new_kwargs['_args'] = new_args
            # inject self as input_map['_task']
            new_kwargs['_task'] = self
            self._set_args(new_args)
            self._set_kwargs(new_kwargs)
            # run the task
            coroutines = [
                asyncio.create_task(
                    self._loop_check(show_done_info=show_done_info)
                ),
                asyncio.create_task(self._run_all(*new_args, **new_kwargs))
            ]
            results = await asyncio.gather(*coroutines)
            result = results[-1]
            self.set_xcom(self.get_name(), f'{result}')
            self._print_result(result)
            return result
        except Exception as e:
            self.log_error(f'{e}')
            if raise_error:
                raise
        finally:
            if show_done_info:
                self._show_env_prefix()
                self._show_run_command()
                self._play_bell()

    async def _loop_check(self, show_done_info: bool = False) -> bool:
        self.log_info('Start readiness checking')
        while not await self._cached_check():
            self.log_debug('Task is not ready')
            await asyncio.sleep(self._checking_interval)
        self._end_timer()
        if show_done_info:
            if show_advertisement:
                selected_advertisement = get_advertisement(advertisements)
                selected_advertisement.show()
            self._show_done_info()
        await self.on_ready()
        return True

    async def _cached_check(self) -> bool:
        if self.__is_check_triggered:
            self.log_debug('Waiting readiness flag to be set')
            while not self.__is_ready:
                await asyncio.sleep(0.1)
            return True
        self.__is_check_triggered = True
        check_result = await self._check()
        if check_result:
            self.__is_ready = True
            self.log_debug('Set readiness flag to True')
        return check_result

    async def _check(self) -> bool:
        '''
        Check current task readiness.
        - If self.checkers is defined,
          this will return True once every self.checkers is completed
        - Otherwise, this will return check method's return value.
        '''
        if len(self._get_checkers()) == 0:
            return await self.check()
        self.log_debug('Waiting execution to be started')
        while not self.__is_execution_started:
            # Don't start checking before the execution itself has been started
            await asyncio.sleep(0.1)
        check_coroutines: Iterable[asyncio.Task] = []
        for checker_task in self._get_checkers():
            checker_task._set_execution_id(self.get_execution_id())
            check_coroutines.append(
                asyncio.create_task(checker_task._run_all())
            )
        await asyncio.gather(*check_coroutines)
        return True

    async def _run_all(self, *args: Any, **kwargs: Any) -> Any:
        await self._mark_awaited()
        coroutines: Iterable[asyncio.Task] = []
        # Add upstream tasks to processes
        self._lock_upstreams()
        for upstream_task in self._get_upstreams():
            upstream_task._set_execution_id(self.get_execution_id())
            coroutines.append(asyncio.create_task(
                upstream_task._run_all(**kwargs)
            ))
        # Add current task to processes
        coroutines.append(self._cached_run(*args, **kwargs))
        # Wait everything to complete
        results = await asyncio.gather(*coroutines)
        if self._return_upstream_result:
            return results[0:-1]
        return results[-1]

    async def _cached_run(self, *args: Any, **kwargs: Any) -> Any:
        if self.__is_execution_triggered:
            self.log_debug('Task has been triggered')
            return
        await self.on_triggered()
        self.__is_execution_triggered = True
        await self.on_waiting()
        # get upstream checker
        upstream_check_processes: Iterable[asyncio.Task] = []
        self._lock_upstreams()
        for upstream_task in self._get_upstreams():
            upstream_check_processes.append(asyncio.create_task(
                upstream_task._loop_check()
            ))
        # wait all upstream checkers to complete
        await asyncio.gather(*upstream_check_processes)
        # mark execution as started, so that checkers can start checking
        self.__is_execution_started = True
        local_kwargs = dict(kwargs)
        local_kwargs['_task'] = self
        if not await self._check_should_execute(*args, **local_kwargs):
            await self.on_skipped()
            await self._mark_done()
            return None
        # start running task
        result: Any = None
        while self._should_attempt():
            try:
                self.log_debug(
                    f'Started with args: {args} and kwargs: {local_kwargs}'
                )
                await self.on_started()
                result = await self.run(*args, **local_kwargs)
                break
            except Exception as e:
                is_last_attempt = self._is_last_attempt()
                await self.on_failed(is_last_attempt, e)
                if is_last_attempt:
                    raise e
                attempt = self._get_attempt()
                self.log_error(f'Encounter error on attempt {attempt}')
                self._increase_attempt()
                await asyncio.sleep(self._retry_interval)
                await self.on_retry()
        await self._mark_done()
        return result

    async def _check_should_execute(self, *args: Any, **kwargs: Any) -> bool:
        if callable(self._should_execute):
            return await run_async(self._should_execute, *args, **kwargs)
        return self.render_bool(self._should_execute)

    async def _set_keyval(self, kwargs: Mapping[str, Any], env_prefix: str):
        # if input is not in input_map, add default values
        for task_input in self._get_combined_inputs():
            key = to_variable_name(task_input.get_name())
            if key in kwargs:
                continue
            kwargs[key] = task_input.get_default()
        # set current task local keyval
        await self._set_local_keyval(kwargs=kwargs, env_prefix=env_prefix)
        # get new_kwargs for upstream and checkers
        new_kwargs = copy.deepcopy(kwargs)
        new_kwargs.update(self.get_input_map())
        upstream_coroutines = []
        # set upstreams keyval
        self._lock_upstreams()
        for upstream_task in self._get_upstreams():
            upstream_coroutines.append(asyncio.create_task(
                upstream_task._set_keyval(
                    kwargs=new_kwargs, env_prefix=env_prefix
                )
            ))
        # set checker keyval
        # local_env_map = self.get_env_map()
        checker_coroutines = []
        for checker_task in self._get_checkers():
            checker_task.add_input(*self._get_inputs())
            checker_task.add_env(*self._get_envs())
            checker_coroutines.append(asyncio.create_task(
                checker_task._set_keyval(
                    kwargs=new_kwargs, env_prefix=env_prefix
                )
            ))
        # wait for checker and upstreams
        coroutines = checker_coroutines + upstream_coroutines
        await asyncio.gather(*coroutines)

    async def _set_local_keyval(
        self, kwargs: Mapping[str, Any], env_prefix: str = ''
    ):
        if self.__is_keyval_set:
            return True
        self.__is_keyval_set = True
        # Set input_map for rendering
        self.log_info('Set input map')
        for task_input in self._get_combined_inputs():
            input_name = to_variable_name(task_input.get_name())
            input_value = kwargs.get(input_name, task_input.get_default())
            if task_input.should_render():
                input_value = self.render_any(input_value)
            self._set_input_map(input_name, input_value)
        self.log_debug(
            'Input map:\n' + map_to_str(self.get_input_map(), item_prefix='  ')
        )
        self.log_info('Merging task envs, task env files, and native envs')
        # Set env_map for rendering
        for env_name, env in self._get_combined_env().items():
            env_value = env.get(env_prefix)
            if env.should_render():
                env_value = self.render_any(env_value)
            self._set_env_map(env_name, env_value)
        self._set_env_map('_ZRB_EXECUTION_ID', self.get_execution_id())
        self.log_debug(
            'Env map:\n' + map_to_str(self.get_env_map(), item_prefix='  ')
        )
        # set task
        self._set_task(self)

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        full_cli_name = double_quote(self._get_full_cli_name())
        return f'<{cls_name} {full_cli_name}>'



from zrb.helper.typing import Any, List
from zrb.helper.python_task import show_lines
from zrb.task.decorator import python_task
from zrb.task_input.str_input import StrInput
from zrb.task_input.int_input import IntInput
from zrb.runner import runner
import datetime
import random

_MIN_WIDTH = 10
_MOTIVATIONAL_QUOTES = [
    [
        'The best time to plant a tree was 20 years ago.',
        'The second best time is now.',
        '~ Chinese Proverb'
    ],
    [
        'The only way to do great work is to love what you do.',
        '~ Steve Jobs'
    ],
    [
        'Believe you can and you\'re halfway there.',
        '~ Theodore Roosevelt'
    ],
    [
        'It does not matter how slowly you go as long as you do not stop.',
        '~ Confucius'
    ],
    [
        'Everything you\'ve ever wanted is on the other side of fear.',
        '~ George Addair'
    ],
    [
        'Success is not final, failure is not fatal:',
        'It is the courage to continue that counts.',
        '~ Winston Churchill'
    ],
    [
        'Hardships often prepare ordinary people',
        'for an extraordinary destiny.',
        '~ C.S. Lewis'
    ],
    [
        'Your time is limited, don\'t waste it living someone else\'s life.',
        '~ Steve Jobs'
    ],
    [
        'Don’t watch the clock; do what it does. Keep going.',
        '~ Sam Levenson'
    ],
    [
        'You are never too old to set another goal or to dream a new dream.',
        '~ C.S. Lewis'
    ],
    [
        'The only limit to our realization of tomorrow',
        'will be our doubts of today.',
        '~ Franklin D. Roosevelt'
    ],
    [
        'Believe in yourself.',
        'You are braver than you think, more talented than you know,'
        'and capable of more than you imagine.',
        '~ Roy T. Bennett'
    ],
    [
        'I can\'t change the direction of the wind,',
        'but I can adjust my sails to always reach my destination.',
        '~ Jimmy Dean'
    ],
    [
        'You are enough just as you are.',
        '~ Meghan Markle'
    ],
    [
        'The future belongs to those',
        'who believe in the beauty of their dreams.',
        '~ Eleanor Roosevelt'
    ]
]


@python_task(
    name='say',
    inputs=[
        StrInput(name='text', default=''),
        IntInput(name='width', default=80)
    ],
    description='Say anything, https://www.youtube.com/watch?v=MbPr1oHO4Hw',
    runner=runner
)
def say(*args: Any, **kwargs: Any):
    width: int = kwargs.get('width', 80)
    if width < _MIN_WIDTH:
        width = _MIN_WIDTH
    text: str = kwargs.get('text', '')
    top_border = '┌' + '─' * (width + 2) + '┐'
    content = [
        '| ' + line + ' |' for line in _get_content(text, width)
    ]
    bottom_border = '└' + '─' * (width + 2) + '┘'
    lines = [top_border] + content + [bottom_border] + [
        '      \\',
        '       \\',
        '  o    ___    o',
        '  | ┌-------┐ |',
        '  |(| o   o |)|',
        '    | └---┘ |',
        '    └-------┘',
    ]
    show_lines(kwargs['_task'], *lines)


def _get_content(text: str, width: int) -> List[str]:
    if text == '':
        now = datetime.datetime.now()
        today = 'Today is ' + now.strftime('%A, %B %d, %Y')
        current_time = 'Current time is ' + now.strftime('%I:%M %p')
        motivational_quote = random.choice(_MOTIVATIONAL_QUOTES)
        return [
            today.ljust(width),
            current_time.ljust(width),
            ''.ljust(width),
        ] + [
            line.ljust(width) for line in motivational_quote
        ]
    return _split_text_by_width(text, width)


def _split_text_by_width(text: str, width: int) -> List[str]:
    original_lines = text.split('\n')
    new_lines = []
    for original_line in original_lines:
        new_lines += [
            original_line[i:i+width].ljust(width)
            for i in range(0, len(original_line), width)
        ]
    return new_lines



from zrb.builtin.generator.project.create import create_project
from zrb.builtin.generator.fastapp.add import add_fastapp
from zrb.builtin.generator.fastapp_module.add import add_fastapp_module
from zrb.builtin.generator.fastapp_crud.add import add_fastapp_crud
import os
import pathlib
import shutil


def test_add_fastapp_crud():
    # prepare path
    dir_path = pathlib.Path(__file__).parent.absolute()
    project_path = os.path.join(dir_path, 'app')
    # remove destination if exists
    if os.path.exists(project_path):
        shutil.rmtree(project_path)

    create_project_fn = create_project.to_function()
    create_project_fn(project_dir=project_path)
    add_fastapp_fn = add_fastapp.to_function()
    add_fastapp_fn(
        project_dir=project_path, app_name='fastapp', http_port=3000
    )
    add_fastapp_module_fn = add_fastapp_module.to_function()
    add_fastapp_module_fn(
        project_dir=project_path, app_name='fastapp', module_name='library'
    )

    # first attempt should success
    first_attempt_fn = add_fastapp_crud.to_function()
    first_attempt_fn(
        project_dir=project_path,
        app_name='fastapp',
        module_name='library',
        entity_name='book',
        plural_entity_name='books',
        column_name='code'
    )

    assert os.path.isdir(os.path.join(
        project_path, 'src', 'fastapp', 'src', 'module', 'library', 'entity', 'book'  # noqa
    ))

    # second attempt should fail
    is_error = False
    try:
        second_attempt_fn = add_fastapp_crud.to_function()
        second_attempt_fn(
            project_dir=project_path,
            app_name='fastapp',
            module_name='library',
            entity_name='book',
            plural_entity_name='books',
            column_name='code'
        )
    except Exception:
        is_error = True
    assert is_error



from zrb.config.config import logging_level
from zrb.helper.typing import Iterable, Mapping, Optional
from zrb.helper.typecheck import typechecked
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.string.parse_replacement import parse_replacement
from zrb.helper.log import logger

import logging
import os
import shutil
import fnmatch


@typechecked
async def copy_tree(
    src: str,
    dst: str,
    replacements: Optional[Mapping[str, str]] = None,
    excludes: Optional[Iterable[str]] = None,
    skip_parsing: Optional[Iterable[str]] = None
):
    if replacements is None:
        replacements = {}
    if excludes is None:
        excludes = []
    if skip_parsing is None:
        skip_parsing = []
    names = os.listdir(src)
    new_dst = parse_replacement(dst, replacements)
    if not os.path.exists(new_dst):
        os.makedirs(new_dst)
    for name in names:
        src_name = os.path.join(src, name)
        if any(fnmatch.fnmatch(src_name, pattern) for pattern in excludes):
            continue
        dst_name = os.path.join(dst, name)
        if os.path.isdir(src_name):
            await copy_tree(
                src_name, dst_name, replacements, excludes, skip_parsing
            )
            continue
        new_dst_name = parse_replacement(dst_name, replacements)
        shutil.copy2(src_name, new_dst_name)
        try:
            if any(
                fnmatch.fnmatch(new_dst_name, pattern)
                for pattern in skip_parsing
            ):
                continue
            file_content = await read_text_file_async(new_dst_name)
            new_file_content = parse_replacement(file_content, replacements)
            await write_text_file_async(new_dst_name, new_file_content)
        except Exception:
            if logging_level <= logging.ERROR:
                logger.error(f'Cannot parse file: {new_dst_name}')



from zrb.task.task import Task
from zrb.task_env.env_file import EnvFile
import os


def test_task_with_env_file():
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('COLOR')
    env_file = os.path.join(
        os.path.dirname(__file__), 'test_task_env_file.env'
    )
    task = Task(
        name='env-file',
        env_files=[EnvFile(path=env_file)],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result == 'blue'


def test_task_with_env_file_and_prefix():
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('COLOR')
    env_file = os.path.join(
        os.path.dirname(__file__), 'test_task_env_file.env'
    )
    task = Task(
        name='env-file-prefixed',
        env_files=[
            EnvFile(
                path=env_file, prefix='ZRB_TEST_TASK_WITH_ENV_AND_PREFIX'
            )
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    if 'ZRB_TEST_TASK_WITH_ENV_AND_PREFIX_COLOR' in os.environ:
        del os.environ['ZRB_TEST_TASK_WITH_ENV_AND_PREFIX_COLOR']
    os.environ['ZRB_TEST_TASK_WITH_ENV_AND_PREFIX_COLOR'] = 'red'
    result = function()
    assert result == 'red'
    del os.environ['ZRB_TEST_TASK_WITH_ENV_AND_PREFIX_COLOR']



from zrb.builtin.generator.project.create import create_project
from zrb.builtin.generator.fastapp.add import add_fastapp
from zrb.builtin.generator.fastapp_module.add import add_fastapp_module
import os
import pathlib
import shutil


def test_add_fastapp_module():
    # prepare path
    dir_path = pathlib.Path(__file__).parent.absolute()
    project_path = os.path.join(dir_path, 'app')
    # remove destination if exists
    if os.path.exists(project_path):
        shutil.rmtree(project_path)

    create_project_fn = create_project.to_function()
    create_project_fn(project_dir=project_path)
    add_fastapp_fn = add_fastapp.to_function()
    add_fastapp_fn(
        project_dir=project_path, app_name='fastapp', http_port=3000
    )

    # first attempt should success
    first_attempt_fn = add_fastapp_module.to_function()
    first_attempt_fn(
        project_dir=project_path, app_name='fastapp', module_name='library'
    )

    # inspect main.py content
    main_py_path = os.path.join(
        project_path, 'src', 'fastapp', 'src', 'main.py'
    )
    with open(main_py_path) as f:
        content = f.read()
    assert 'from component.app import app' in content
    assert 'from module.library.register_module import register_library' in content  # noqa
    assert 'register_library()' in content

    # second attempt should fail
    is_error = False
    try:
        second_attempt_fn = add_fastapp_module.to_function()
        second_attempt_fn(
            project_dir=project_path, app_name='fastapp', module_name='library'
        )
    except Exception:
        is_error = True
    assert is_error



from zrb.helper.typing import (
    Any, Callable, Iterable, List, Mapping, Optional, Union
)
from zrb.helper.typecheck import typechecked
from zrb.task.any_task_event_handler import (
    OnTriggered, OnWaiting, OnSkipped, OnStarted, OnReady, OnRetry, OnFailed
)
from zrb.task.any_task import AnyTask
from zrb.helper.string.conversion import to_cli_name
from zrb.helper.accessories.color import get_random_color
from zrb.helper.accessories.icon import get_random_icon
from zrb.helper.util import coalesce_str
from zrb.task_input.any_input import AnyInput
from zrb.task_group.group import Group
from zrb.task_env.constant import RESERVED_ENV_NAMES
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile

import os


@typechecked
class CommonTaskModel():
    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        description: str = '',
        inputs: List[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0,
        run: Optional[Callable[..., Any]] = None,
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
        return_upstream_result: bool = False
    ):
        self._name = name
        self._group = group
        if group is not None:
            group._add_task(self)
        self._description = coalesce_str(description, name)
        self._inputs = inputs
        self._envs = envs
        self._env_files = env_files
        self._icon = coalesce_str(icon, get_random_icon())
        self._color = coalesce_str(color, get_random_color())
        self._retry = retry
        self._retry_interval = retry_interval
        self._upstreams = upstreams
        self._checkers = [checker.copy() for checker in checkers]
        self._checking_interval = checking_interval
        self._run_function: Optional[Callable[..., Any]] = run
        self._on_triggered = on_triggered
        self._on_waiting = on_waiting
        self._on_skipped = on_skipped
        self._on_started = on_started
        self._on_ready = on_ready
        self._on_retry = on_retry
        self._on_failed = on_failed
        self._should_execute = should_execute
        self._return_upstream_result = return_upstream_result
        self.__execution_id = ''
        self.__allow_add_envs = True
        self.__allow_add_env_files = True
        self.__allow_add_inputs = True
        self.__allow_add_upstreams: bool = True
        self.__allow_add_checkers: bool = True
        self.__has_already_inject_env_files: bool = False
        self.__has_already_inject_envs: bool = False
        self.__has_already_inject_inputs: bool = False
        self.__has_already_inject_upstreams: bool = False
        self.__all_inputs: Optional[List[AnyInput]] = None

    def _lock_upstreams(self):
        self.__allow_add_upstreams = False

    def _set_execution_id(self, execution_id: str):
        if self.__execution_id == '':
            self.__execution_id = execution_id

    def _propagate_execution_id(self):
        execution_id = self.get_execution_id()
        for upstream_task in self._get_upstreams():
            upstream_task._set_execution_id(execution_id)
            upstream_task._propagate_execution_id()
        for checker_task in self._get_checkers():
            checker_task._set_execution_id(execution_id)
            checker_task._propagate_execution_id()

    def get_execution_id(self) -> str:
        return self.__execution_id

    def set_name(self, new_name: str):
        if self._description == self.get_name():
            self._description = new_name
        self._name = new_name

    def get_name(self) -> str:
        return self._name

    def get_cli_name(self) -> str:
        return to_cli_name(self.get_name())

    def set_description(self, new_description: str):
        self._description = new_description

    def get_description(self) -> str:
        return self._description

    def set_icon(self, new_icon: str):
        self._icon = new_icon

    def set_color(self, new_color: str):
        self._color = new_color

    def set_retry(self, new_retry: int):
        self._retry = new_retry

    def set_should_execute(
        self, should_execute: Union[bool, str, Callable[..., bool]]
    ):
        self._should_execute = should_execute

    def set_retry_interval(self, new_retry_interval: Union[float, int]):
        self._retry_interval = new_retry_interval

    def set_checking_interval(self, new_checking_interval: Union[float, int]):
        self._checking_interval = new_checking_interval

    def insert_input(self, *inputs: AnyInput):
        if not self.__allow_add_inputs:
            raise Exception(f'Cannot insert inputs for `{self.get_name()}`')
        self._inputs = list(inputs) + list(self._inputs)

    def add_input(self, *inputs: AnyInput):
        if not self.__allow_add_inputs:
            raise Exception(f'Cannot add inputs for `{self.get_name()}`')
        self._inputs = list(self._inputs) + list(inputs)

    def inject_inputs(self):
        pass

    def _get_inputs(self) -> List[AnyInput]:
        if not self.__has_already_inject_inputs:
            self.inject_inputs()
            self.__has_already_inject_inputs = True
        return list(self._inputs)

    def _get_combined_inputs(self) -> Iterable[AnyInput]:
        ''''
        Getting all inputs of this task and all its upstream, non-duplicated.
        '''
        if self.__all_inputs is not None:
            return self.__all_inputs
        self.__all_inputs: List[AnyInput] = []
        existing_input_names: Mapping[str, bool] = {}
        # Add task inputs
        inputs = self._get_inputs()
        for input_index, first_occurence_task_input in enumerate(inputs):
            input_name = first_occurence_task_input.get_name()
            if input_name in existing_input_names:
                continue
            # Look for all input with the same name in the current task
            task_inputs = [
                candidate
                for candidate in inputs[input_index:]
                if candidate.get_name() == input_name
            ]
            # Get the last input, and add it to _all_inputs
            task_input = task_inputs[-1]
            self.__all_inputs.append(task_input)
            existing_input_names[input_name] = True
        # Add upstream inputs
        for upstream in self._get_upstreams():
            upstream_inputs = upstream._get_combined_inputs()
            for upstream_input in upstream_inputs:
                if upstream_input.get_name() in existing_input_names:
                    continue
                self.__all_inputs.append(upstream_input)
                existing_input_names[upstream_input.get_name()] = True
        self._lock_upstreams()
        self.__allow_add_inputs = False
        return self.__all_inputs

    def insert_env(self, *envs: Env):
        if not self.__allow_add_envs:
            raise Exception(f'Cannot insert envs to `{self.get_name()}`')
        self._envs = list(envs) + list(self._envs)

    def add_env(self, *envs: Env):
        if not self.__allow_add_envs:
            raise Exception(f'Cannot add envs to `{self.get_name()}`')
        self._envs = list(self._envs) + list(envs)

    def inject_envs(self):
        pass

    def _get_envs(self) -> List[Env]:
        if not self.__has_already_inject_envs:
            self.inject_envs()
            self.__has_already_inject_envs = True
        return list(self._envs)

    def _get_combined_env(self) -> Mapping[str, Env]:
        all_envs: Mapping[str, Env] = {}
        for env_name in os.environ:
            if env_name in RESERVED_ENV_NAMES:
                continue
            all_envs[env_name] = Env(
                name=env_name, os_name=env_name, should_render=False
            )
        for env_file in self._get_env_files():
            for env in env_file.get_envs():
                all_envs[env.get_name()] = env
        for env in self._get_envs():
            all_envs[env.get_name()] = env
        self.__allow_add_envs = False
        self.__allow_add_env_files = False
        return all_envs

    def insert_env_file(self, *env_files: EnvFile):
        if not self.__allow_add_env_files:
            raise Exception(f'Cannot insert env_files to `{self.get_name()}`')
        self._env_files = list(env_files) + list(self._env_files)

    def add_env_file(self, *env_files: EnvFile):
        if not self.__allow_add_env_files:
            raise Exception(f'Cannot add env_files to `{self.get_name()}`')
        self._env_files = list(self._env_files) + list(env_files)

    def inject_env_files(self):
        pass

    def insert_upstream(self, *upstreams: AnyTask):
        if not self.__allow_add_upstreams:
            raise Exception(f'Cannot insert upstreams to `{self.get_name()}`')
        self._upstreams = list(upstreams) + list(self._upstreams)

    def add_upstream(self, *upstreams: AnyTask):
        if not self.__allow_add_upstreams:
            raise Exception(f'Cannot add upstreams to `{self.get_name()}`')
        self._upstreams = list(self._upstreams) + list(upstreams)

    def inject_upstreams(self):
        pass

    def _get_upstreams(self) -> List[AnyTask]:
        if not self.__has_already_inject_upstreams:
            self.inject_upstreams()
            self.__has_already_inject_upstreams = True
        return list(self._upstreams)

    def get_icon(self) -> str:
        return self._icon

    def get_color(self) -> str:
        return self._color

    def _get_env_files(self) -> List[EnvFile]:
        if not self.__has_already_inject_env_files:
            self.inject_env_files()
            self.__has_already_inject_env_files = True
        return self._env_files

    def insert_checker(self, *checkers: AnyTask):
        if not self.__allow_add_checkers:
            raise Exception(f'Cannot insert checkers to `{self.get_name()}`')
        additional_checkers = [checker.copy() for checker in checkers]
        self._checkers = additional_checkers + self._checkers

    def add_checker(self, *checkers: AnyTask):
        if not self.__allow_add_checkers:
            raise Exception(f'Cannot add checkers to `{self.get_name()}`')
        additional_checkers = [checker.copy() for checker in checkers]
        self._checkers = self._checkers + additional_checkers

    def inject_checkers(self):
        pass

    def _get_checkers(self) -> List[AnyTask]:
        if not self.__allow_add_checkers:
            self.inject_checkers()
            self.__allow_add_checkers = True
        return self._checkers



from zrb import PortChecker

snake_zrb_app_name_checker = PortChecker(
    name='check-kebab-zrb-app-name',
    host='{{input.snake_zrb_app_name_host}}',
    port='{{env.APP_PORT}}'
)

snake_zrb_app_name_container_checker = PortChecker(
    name='check-kebab-zrb-app-name',
    host='{{input.snake_zrb_app_name_host}}',
    port='{{env.HOST_PORT}}'
)



from zrb.helper.typing import Any
from zrb.task.task import Task
from zrb.task.decorator import python_task
from zrb.task.cmd_task import CmdTask
from zrb.task.notifier import Notifier
from zrb.task_input.str_input import StrInput


def create_recurring_action(
    notif_title: str,
    trigger_caption: str,
    trigger_xcom_key: str,
    default_message: str = '👋'
) -> Task:
    # define inputs
    message_input = StrInput(
        name='message',
        default=default_message,
        prompt='Message to be shown',
    )
    command_input = StrInput(
        name='command',
        default='',
        prompt='Command to be executed',
    )
    # define tasks
    show_trigger_info = _create_show_trigger_info(
        trigger_caption=trigger_caption,
        trigger_xcom_key=trigger_xcom_key
    )
    run_command = CmdTask(
        name='run-command',
        icon='⚙️',
        color='blue',
        inputs=[command_input],
        upstreams=[show_trigger_info],
        should_execute='{{ input.command != "" }}',
        cmd='{{ input.command }}'
    )
    notify = Notifier(
        name='notify',
        icon='📢',
        color='green',
        inputs=[message_input],
        title=notif_title,
        message='{{ input.message }}',
        upstreams=[show_trigger_info],
        should_execute='{{ input.message != "" }}',
    )
    # return aggregator task
    return Task(
        name='recurring-action',
        inputs=[message_input, command_input],
        upstreams=[run_command, notify],
        retry=0
    )


def _create_show_trigger_info(
    trigger_caption: str,
    trigger_xcom_key: str
) -> Task:
    @python_task(
        name='show-trigger-info',
        icon='🔍',
        color='magenta',
    )
    def show_trigger_info(*args: Any, **kwargs: Any):
        task: Task = kwargs.get('_task')
        task.print_out(f'{trigger_caption}: {task.get_xcom(trigger_xcom_key)}')
    return show_trigger_info



from zrb.task.task import Task
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
import os


def test_task_env_with_jinja_value():
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('ZRB_TEST_TASK_ENV_2')
    task = Task(
        name='templated-env',
        envs=[
            Env(name='ZRB_TEST_TASK_ENV_1', default='Elixir'),
            Env(
                name='ZRB_TEST_TASK_ENV_2',
                default='{{env.ZRB_TEST_TASK_ENV_1}} of immortality'
            )
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    if 'ZRB_TEST_TASK_ENV_1' in os.environ:
        del os.environ['ZRB_TEST_TASK_ENV_1']
    if 'ZRB_TEST_TASK_ENV_2' in os.environ:
        del os.environ['ZRB_TEST_TASK_ENV_2']
    result = function()
    assert result == 'Elixir of immortality'


def test_task_env_with_should_not_be_rendered_jinja_value():
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('ZRB_TEST_TASK_ENV_2')
    task = Task(
        name='templated-env',
        envs=[
            Env(name='ZRB_TEST_TASK_ENV_1', default='Elixir'),
            Env(
                name='ZRB_TEST_TASK_ENV_2',
                default='{{env.ZRB_TEST_TASK_ENV_1}} of immortality',
                should_render=False
            )
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    if 'ZRB_TEST_TASK_ENV_1' in os.environ:
        del os.environ['ZRB_TEST_TASK_ENV_1']
    if 'ZRB_TEST_TASK_ENV_2' in os.environ:
        del os.environ['ZRB_TEST_TASK_ENV_2']
    result = function()
    assert result == '{{env.ZRB_TEST_TASK_ENV_1}} of immortality'


def test_task_env_with_none_as_os_name():
    '''
    When env exist, it should override env_file
    '''
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('COLOR')
    env_file = os.path.join(
        os.path.dirname(__file__), 'test_task_env_file.env'
    )
    task = Task(
        name='env-file-prefixed',
        env_files=[
            EnvFile(path=env_file, prefix='ZRB_TEST_TASK_OS_NAME_NONE')
        ],
        envs=[Env(name='COLOR', os_name=None, default='green')],
        run=_run,
        retry=0
    )
    function = task.to_function()
    if 'ZRB_TEST_TASK_OS_NAME_NONE_COLOR' in os.environ:
        del os.environ['ZRB_TEST_TASK_OS_NAME_NONE_COLOR']
    if 'COLOR' in os.environ:
        del os.environ['COLOR']
    os.environ['ZRB_TEST_TASK_OS_NAME_NONE_COLOR'] = 'red'
    os.environ['COLOR'] = 'cyan'
    result = function()
    assert result == 'cyan'
    del os.environ['ZRB_TEST_TASK_OS_NAME_NONE_COLOR']
    del os.environ['COLOR']


def test_task_env_with_empty_string_as_os_name():
    '''
    When env exist, it should override env_file,
    If env has empty os_name, the default value should always be used.
    '''
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('COLOR')
    env_file = os.path.join(
        os.path.dirname(__file__), 'test_task_env_file.env'
    )
    task = Task(
        name='env-file-prefixed',
        env_files=[
            EnvFile(path=env_file, prefix='ZRB_TEST_TASK_OS_NAME_EMPTY')
        ],
        envs=[Env(name='COLOR', os_name='', default='green')],
        run=_run,
        retry=0
    )
    function = task.to_function()
    if 'ZRB_TEST_TASK_OS_NAME_EMPTY_COLOR' in os.environ:
        del os.environ['ZRB_TEST_TASK_OS_NAME_EMPTY_COLOR']
    if 'COLOR' in os.environ:
        del os.environ['COLOR']
    os.environ['ZRB_TEST_TASK_OS_NAME_EMPTY_COLOR'] = 'red'
    os.environ['COLOR'] = 'cyan'
    result = function()
    assert result == 'green'
    del os.environ['ZRB_TEST_TASK_OS_NAME_EMPTY_COLOR']
    del os.environ['COLOR']


def test_task_duplicate_env():
    '''
    When env exist, it should override env_file,
    If two env has the same name, the later should override the first one
    '''
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('COLOR')
    env_file = os.path.join(
        os.path.dirname(__file__), 'test_task_env_file.env'
    )
    task = Task(
        name='env-file-prefixed',
        env_files=[EnvFile(path=env_file, prefix='ZRB_TEST_TASK')],
        envs=[
            Env(name='COLOR', os_name='', default='green'),
            Env(name='COLOR', os_name='', default='yellow')
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result == 'yellow'



from zrb.task.task import Task
from zrb.task_input.str_input import StrInput
import asyncio


def test_task_with_no_runner():
    task = Task(
        name='simple-no-error',
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result is None


def test_task_with_predefined_runner():
    def _run(*args, **kwargs) -> str:
        return 'hello'
    task = Task(
        name='hello',
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result == 'hello'


def test_should_not_executed_task():
    def _run(*args, **kwargs) -> str:
        return 'hello'
    task = Task(
        name='hello',
        run=_run,
        retry=0,
        should_execute=False
    )
    function = task.to_function()
    result = function()
    assert result is None


def test_task_returning_upstream_result():
    task_upstream_1 = Task(
        name='task-upstream-1',
        run=lambda *args, **kwargs: 'articuno'
    )
    task_upstream_2 = Task(
        name='task-upstream-2',
        run=lambda *args, **kwargs: 'zapdos'
    )
    task_upstream_3 = Task(
        name='task-upstream-3',
        run=lambda *args, **kwargs: 'moltres'
    )
    task = Task(
        name='task',
        upstreams=[
            task_upstream_1, task_upstream_2, task_upstream_3
        ],
        return_upstream_result=True
    )
    function = task.to_function()
    result = function()
    assert len(result) == 3
    assert 'articuno' in result
    assert 'zapdos' in result
    assert 'moltres' in result


def test_task_to_async_function():
    task = Task(
        name='task',
        inputs=[
            StrInput(name='name', default='zapdos')
        ],
        run=lambda *args, **kwargs: kwargs.get('name')
    )
    function = task.to_function(is_async=True)
    result = asyncio.run(function())
    assert result == 'zapdos'


def test_task_function_should_accept_kwargs_args():
    task = Task(
        name='task',
        run=lambda *args, **kwargs: args[0]
    )
    function = task.to_function()
    # _args keyword argument should be treated as *args
    result = function(_args=['moltres'])
    assert result == 'moltres'


def test_callable_as_task_should_execute_value():
    # should execute should accept executable
    task = Task(
        name='task',
        should_execute=lambda *args, **kwargs: True,
        run=lambda *args, **kwargs: 'articuno'
    )
    function = task.to_function()
    result = function()
    assert result == 'articuno'



from zrb.task.recurring_task import RecurringTask
from zrb.task.time_watcher import TimeWatcher
from zrb.task_input.str_input import StrInput
from zrb.runner import runner
from zrb.builtin.helper.reccuring_action import create_recurring_action


schedule = RecurringTask(
    name='schedule',
    icon='📅',
    color='yellow',
    description='Show message/run command periodically',
    inputs=[
        StrInput(
            name='schedule',
            default='* * * * *',
            prompt='Schedule cron pattern (minute hour day(month) month day(week)',  # noqa
            description='Schedule cron pattern to show the message'
        ),
    ],
    triggers=[
        TimeWatcher(
            name='watch-time',
            color='cyan',
            icon='⏰',
            schedule='{{input.schedule}}'
        )
    ],
    task=create_recurring_action(
        notif_title='Schedule',
        trigger_caption='Schedule',
        trigger_xcom_key='watch-time.scheduled-time'
    )
)
runner.register(schedule)



from logging import Logger
from core.messagebus import Publisher, Consumer
from core.rpc import Caller
from module.log.entity.activity.event import (
    register_event as register_activity_event
)


def register_event(
    logger: Logger,
    consumer: Consumer,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register event handlers for "log"')
    register_activity_event(logger, consumer, rpc_caller, publisher)



from zrb.helper.typecheck import typechecked
from zrb.runner import runner
from zrb.config.config import (
    init_scripts, should_load_builtin, version, logging_level
)
from zrb.helper.loader.load_module import load_module
from zrb.helper.log import logger
from zrb.helper.accessories.color import colored
from functools import lru_cache

import click
import logging
import os
import sys

HELP = f'''
                bb
   zzzzz rr rr  bb
     zz  rrr  r bbbbbb
    zz   rr     bb   bb
   zzzzz rr     bbbbbb   {version}
   _ _ . .  . _ .  _ . . .

Super framework for your super app.

☕ Donate at: https://stalchmst.com/donation
🐙 Submit issues/PR at: https://github.com/state-alchemists/zrb
🐤 Follow us at: https://twitter.com/zarubastalchmst
'''


class MultilineHelpClickGroup(click.Group):
    def format_help_text(self, ctx, formatter):
        formatter.write(self.help)


@lru_cache
@typechecked
def create_cli() -> click.Group:
    if logging_level <= logging.INFO:
        logger.info(colored('Prepare CLI', attrs=['dark']))
    zrb_cli_group = MultilineHelpClickGroup(name='zrb', help=HELP)
    # load from ZRB_INIT_SCRIPTS environment
    for init_script in init_scripts:
        if logging_level <= logging.INFO:
            logger.info(colored(
                f'Load modules from {init_script}', attrs=['dark']
            ))
        load_module(script_path=init_script)
    # Load default tasks
    if should_load_builtin:
        if logging_level <= logging.INFO:
            logger.info(colored('Load builtins', attrs=['dark']))
        from zrb import builtin
        assert builtin
    # Load zrb_init.py
    project_dir = os.getenv('ZRB_PROJECT_DIR', os.getcwd())
    _load_zrb_init(project_dir)
    # Serve all tasks registered to runner
    if logging_level <= logging.INFO:
        logger.info(colored('Serve CLI', attrs=['dark']))
    cli = runner.serve(zrb_cli_group)
    return cli


@lru_cache
@typechecked
def _load_zrb_init(project_dir: str):
    project_script = os.path.join(project_dir, 'zrb_init.py')
    if not os.path.isfile(project_script):
        return
    sys.path.append(project_dir)
    if logging_level <= logging.INFO:
        logger.info(colored(f'Set sys.path to {sys.path}', attrs=['dark']))
    python_path = _get_new_python_path(project_dir)
    if logging_level <= logging.INFO:
        logger.info(
            colored(f'Set PYTHONPATH to {python_path}', attrs=['dark'])
        )
    os.environ['PYTHONPATH'] = python_path
    if logging_level <= logging.INFO:
        logger.info(colored(f'Load modules from {project_script}', attrs=['dark']))
    load_module(script_path=project_script)


@typechecked
def _get_new_python_path(project_dir: str) -> str:
    current_python_path = os.getenv('PYTHONPATH')
    if current_python_path is None or current_python_path == '':
        return project_dir
    return ':'.join([current_python_path, project_dir])



from zrb.builtin.group import ubuntu_group, ubuntu_install_group
from zrb.task.cmd_task import CmdTask
from zrb.task.flow_task import FlowTask
from zrb.runner import runner
from zrb.builtin.devtool.devtool_install import (
    install_tmux, install_zsh, install_gvm, install_nvm, install_pyenv,
    install_sdkman, install_aws, install_gcloud, install_docker,
    install_kubectl, install_helm, install_helix
)

###############################################################################
# Task Definitions
###############################################################################


update = CmdTask(
    name='update',
    group=ubuntu_group,
    description='Update ubuntu',
    cmd=[
        'sudo apt update',
        'sudo apt upgrade -y',
    ],
    retry_interval=3,
    preexec_fn=None
)
runner.register(update)

install_toys = CmdTask(
    name='toys',
    group=ubuntu_install_group,
    description='Install ubuntu toy packages',
    cmd=[
        'sudo apt install -y lolcat cowsay figlet neofetch',
    ],
    retry_interval=3,
    preexec_fn=None
)
update_and_install_toys: CmdTask = install_toys.copy()
update_and_install_toys.add_upstream(update)
runner.register(update_and_install_toys)

install_essentials = CmdTask(
    name='essentials',
    group=ubuntu_install_group,
    description='Install ubuntu essential packages',
    cmd=[
        'sudo apt install -y \\',
        'build-essential python3-distutils libssl-dev zlib1g-dev \\'
        'libbz2-dev libreadline-dev libsqlite3-dev libpq-dev python3-dev \\',
        'llvm libncurses5-dev libncursesw5-dev xz-utils tk-dev libffi-dev \\',
        'liblzma-dev python3-openssl libblas-dev liblapack-dev rustc \\',
        'golang gfortran fd-find ripgrep wget curl git ncat zip unzip \\',
        'cmake make tree tmux zsh neovim xdotool xsel'
    ],
    retry_interval=3,
    preexec_fn=None
)
update_and_install_essentials: CmdTask = install_essentials.copy()
update_and_install_essentials.add_upstream(update)
runner.register(update_and_install_essentials)

install_tex = CmdTask(
    name='tex',
    group=ubuntu_install_group,
    description='Install ubuntu tex packages',
    cmd=[
        'sudo apt install -y \\',
        'texlive-full texlive-latex-base texlive-fonts-recommended \\',
        'texlive-fonts-extra texlive-latex-extra'
    ],
    retry_interval=3,
    preexec_fn=None
)
update_and_install_tex: CmdTask = install_tex.copy()
update_and_install_tex.add_upstream(update)
runner.register(update_and_install_tex)

install_all = FlowTask(
    name='all',
    group=ubuntu_install_group,
    description='Install all ubuntu packages',
    steps=[
        update,
        install_essentials,
        install_toys,
        install_tex,
        install_zsh,
        install_tmux,
        install_gvm,
        install_nvm,
        install_pyenv,
        install_sdkman,
        install_aws,
        install_gcloud,
        install_docker,
        install_kubectl,
        install_helm,
        install_helix
    ]
)
runner.register(install_all)



from zrb.helper.codemod.add_upstream_to_task import add_upstream_to_task


def test_add_upstream_to_task_with_upstream():
    code = add_upstream_to_task(
        code='\n'.join([
            "a = Task(name='alpha', upstreams=[])",
            "b = Task(name='beta', upstreams=[a])",
            "c = Task(name='gamma', upstreams=[b])",
        ]),
        task_name='beta',
        upstream_task_name='delta'
    )
    assert code == '\n'.join([
        "a = Task(name='alpha', upstreams=[])",
        "b = Task(name='beta', upstreams=[a, delta])",
        "c = Task(name='gamma', upstreams=[b])",
    ])


def test_add_upstream_to_task_without_upstream():
    code = add_upstream_to_task(
        code='\n'.join([
            "a = Task(name='alpha')",
            "b = Task(name='beta')",
            "c = Task(name='gamma')",
        ]),
        task_name='beta',
        upstream_task_name='delta'
    )
    assert code == '\n'.join([
        "a = Task(name='alpha')",
        "b = Task(name='beta', upstreams = [delta])",
        "c = Task(name='gamma')",
    ])



from zrb.helper.typing import Any, Callable, Iterable, Optional, Union
from zrb.helper.typecheck import typechecked
from zrb.task.base_task.base_task import BaseTask
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnTriggered, OnWaiting, OnSkipped, OnStarted, OnReady, OnRetry, OnFailed
)
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

import asyncio


@typechecked
class Checker(BaseTask):

    def __init__(
        self,
        name: str = 'check',
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        upstreams: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checking_interval: Union[int, float] = 0.1,
        progress_interval: Union[int, float] = 30,
        expected_result: bool = True,
        should_execute: Union[bool, str, Callable[..., bool]] = True
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=[],
            checking_interval=checking_interval,
            retry=0,
            retry_interval=0,
            should_execute=should_execute,
        )
        self._progress_interval = progress_interval
        self._expected_result = expected_result
        self._should_show_progress = False

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        wait_time = 0
        while True:
            self._should_show_progress = wait_time >= self._progress_interval
            inspect_result = await self.inspect(*args, **kwargs)
            if inspect_result == self._expected_result:
                return True
            if wait_time >= self._progress_interval:
                wait_time = 0
            await asyncio.sleep(self._checking_interval)
            wait_time += self._checking_interval

    async def inspect(self, *args: Any, **kwargs: Any) -> bool:
        return False

    def show_progress(self, message: str):
        if self._should_show_progress:
            self.print_out_dark(message)
            return
        self.log_debug(message)



from zrb.helper.typecheck import typechecked
import random
import string


@typechecked
def get_random_name(
    separator: str = '-',
    add_random_digit: bool = True,
    digit_count: int = 4
) -> str:
    prefixes = [
        'albedo', 'argent', 'argentum', 'aurora', 'aurum', 'azure',
        'basilisk', 'cerulean', 'chimeric', 'citrin', 'coral', 'crimson',
        'diamond', 'draco', 'dragon', 'emerald', 'ethereal', 'ferrum',
        'flammeus', 'garnet', 'glacial', 'glimmering', 'glistening', 'golden',
        'helios', 'igneous', 'imperial', 'jade', 'luminous', 'luna', 'lunar',
        'mystic', 'nephrite', 'nocturnal', 'obsidian', 'opal', 'pearl',
        'platinum', 'prismatic', 'ruby', 'sapphire', 'serpentine', 'silver',
        'sol', 'solar', 'spiritual', 'stellar',  'tempest', 'topaz',
        'turquoise', 'verde', 'vermillion', 'vitreous', 'zephyr', 'zircon'
    ]
    suffixes = [
        'aether', 'albedo', 'alchemy', 'arcana', 'aureum', 'aetheris',
        'anima', 'astralis', 'caelestis', 'chrysopoeia', 'cosmicum',
        'crystallum', 'deum', 'divinitas', 'draconis', 'elementorum', 'elixir',
        'essentia', 'eternis', 'ethereus', 'fatum', 'flamma', 'fulgur',
        'hermetica', 'ignis', 'illuminationis', 'imperium', 'incantatum',
        'infinitum', 'lapis', 'lux', 'magicae', 'magnum', 'materia',
        'metallum', 'mysticum', 'natura', 'occultum', 'omnipotentis',
        'opulentia', 'philosophia', 'philosophorum', 'praeparatum',
        'praestantissimum', 'prima', 'primordium', 'quintessentia',
        'regeneratio', 'ritualis', 'sanctum', 'spiritus', 'tenebris',
        'terra', 'tinctura', 'transmutationis', 'universalis', 'vapores',
        'venenum', 'veritas', 'vitae', 'volatus'
    ]
    prefix = random.choice(prefixes)
    suffix = random.choice(suffixes)
    parts = [prefix, suffix]
    if add_random_digit:
        random_digit = ''.join(
            random.choices(string.digits, k=digit_count)
        )
        parts.append(random_digit)
    return separator.join(parts)



from typing import List
from helper.conversion import str_to_boolean, str_to_logging_level
import os
import json

app_src_dir = os.path.dirname(__file__)

zrb_app_name = os.getenv('APP_NAME', 'app')
app_logging_level = str_to_logging_level(
    os.getenv('APP_LOGGING_LEVEL', 'INFO')
)
app_host = os.getenv('APP_HOST', '0.0.0.0')
app_port = int(os.getenv('APP_PORT', '8080'))
app_reload = str_to_boolean(os.getenv('APP_RELOAD', 'true'))
app_max_not_ready = int(os.getenv('APP_MAX_NOT_READY', '10'))

app_enable_otel = str_to_boolean(os.getenv('APP_ENABLE_OTEL', 'false'))
app_otel_exporter_otlp_endpoint = os.getenv(
    'APP_OTEL_EXPORTER_OTLP_ENDPOINT', 'false'
)

app_auth_access_token_cookie_key = os.getenv(
    'PUBLIC_AUTH_ACCESS_TOKEN_COOKIE_KEY', 'access_token'
)
app_auth_refresh_token_cookie_key = os.getenv(
    'PUBLIC_AUTH_REFRESH_TOKEN_COOKIE_KEY', 'refresh_token'
)

app_auth_access_token_type = os.getenv('APP_AUTH_ACCESS_TOKEN_TYPE', 'jwt')
app_auth_access_token_expire_seconds = int(os.getenv(
    'APP_AUTH_ACCESS_TOKEN_EXPIRE_SECONDS', '300'
))
app_auth_refresh_token_type = os.getenv('APP_AUTH_REFRESH_TOKEN_TYPE', 'jwt')
app_auth_refresh_token_expire_seconds = int(os.getenv(
    'APP_AUTH_REFRESH_TOKEN_EXPIRE_SECONDS', '86400'
))
app_auth_jwt_token_secret_key = os.getenv(
    'APP_AUTH_JWT_TOKEN_SECRET_KEY', 'secret'
)
app_auth_jwt_token_algorithm = os.getenv(
    'APP_AUTH_JWT_TOKEN_ALGORITHM', 'HS256'
)

app_auth_admin_active = str_to_boolean(os.getenv(
    'APP_AUTH_ADMIN_ACTIVE', 'true'
))
app_auth_admin_user_id = os.getenv('APP_AUTH_ADMIN_USER_ID', 'root')
app_auth_admin_username = os.getenv('APP_AUTH_ADMIN_USERNAME', 'admin')
app_auth_admin_password = os.getenv('APP_AUTH_ADMIN_PASSWORD', 'admin')
app_auth_admin_email = os.getenv('APP_AUTH_ADMIN_EMAIL', '')
app_auth_admin_phone = os.getenv('APP_AUTH_ADMIN_PHONE', '')
app_auth_guest_user_id = os.getenv('APP_AUTH_GUEST_USER_ID', 'guest')
app_auth_guest_username = os.getenv('APP_AUTH_GUEST_USERNAME', 'guest')
app_auth_guest_email = os.getenv('APP_AUTH_GUEST_EMAIL', '')
app_auth_guest_phone = os.getenv('APP_AUTH_GUEST_PHONE', '')

app_db_connection = os.getenv('APP_DB_CONNECTION', 'sqlite://')
app_db_engine_show_log = str_to_boolean(
    os.getenv('APP_DB_ENGINE_SHOW_LOG', 'false')
)
app_db_auto_migrate = str_to_boolean(os.getenv('APP_DB_AUTO_MIGRATE', 'true'))

app_broker_type = os.getenv('APP_BROKER_TYPE', 'mock')

app_enable_rpc_server: bool = str_to_boolean(os.getenv(
    'APP_ENABLE_RPC_SERVER', 'true'
))

app_enable_event_handler: bool = str_to_boolean(os.getenv(
    'APP_ENABLE_EVENT_HANDLER', 'true'
))

app_enable_frontend: bool = str_to_boolean(os.getenv(
    'APP_ENABLE_FRONTEND', 'true'
))

app_enable_api: bool = str_to_boolean(os.getenv(
    'APP_ENABLE_API', 'true'
))

app_rmq_connection_string = os.getenv(
    'APP_RMQ_CONNECTION', 'amqp://guest:guest@localhost/'
)

app_kafka_bootstrap_servers = os.getenv(
    'APP_KAFKA_BOOTSTRAP_SERVERS', 'localhost:9092'
)
app_kafka_security_protocol = os.getenv(
    'APP_KAFKA_SECURITY_PROTOCOL', 'PLAINTEXT'
)
app_kafka_sasl_mechanism = os.getenv(
    'APP_KAFKA_SASL_MECHANISM', 'SCRAM-SHA-512'
)
app_kafka_sasl_user = os.getenv(
    'APP_KAFKA_SASL_USER', 'admin'
)
app_kafka_sasl_pass = os.getenv(
    'APP_KAFKA_SASL_PASS', 'admin'
)

app_cors_allow_origins: List[str] = json.loads(os.getenv(
    'APP_CORS_ALLOW_ORIGINS', '["*"]'
))
app_cors_allow_origin_regex: str = os.getenv(
    'APP_CORS_ALLOW_ORIGIN_REGEX', ''
)
app_cors_allow_methods: List[str] = json.loads(os.getenv(
    'APP_CORS_ALLOW_METHODS', '["*"]'
))
app_cors_allow_headers: List[str] = json.loads(os.getenv(
    'APP_CORS_ALLOW_HEADERS', '["*"]'
))
app_cors_allow_credentials: bool = str_to_boolean(os.getenv(
    'APP_CORS_ALLOW_CREDENTIALS', 'false'
))
app_cors_expose_headers: bool = str_to_boolean(os.getenv(
    'APP_CORS_EXPOSE_HEADERS', 'false'
))
app_cors_max_age: int = int(os.getenv(
    'APP_CORS_MAX_AGE', '600'
))
app_enable_auth_module = str_to_boolean(os.getenv(
    'APP_ENABLE_AUTH_MODULE', 'true'
))
app_enable_log_module = str_to_boolean(os.getenv(
    'APP_ENABLE_LOG_MODULE', 'true'
))

public_brand = os.getenv('PUBLIC_BRAND', 'PascalZrbAppName')
public_title = os.getenv('PUBLIC_TITLE', 'PascalZrbAppName')
public_auth_access_token_cookie_key = os.getenv(
    'PUBLIC_AUTH_ACCESS_TOKEN_COOKIE_KEY', 'access_token'
)
public_auth_refresh_token_cookie_key = os.getenv(
    'PUBLIC_AUTH_REFRESH_TOKEN_COOKIE_KEY', 'refresh_token'
)



from zrb.helper.typing import Any
from zrb.builtin.generator.common.task_input import project_dir_input
from zrb.builtin.generator.common.helper import (
    validate_existing_project_dir, validate_inexisting_automation
)
from zrb.builtin.generator.common.task_factory import create_register_module
from zrb.builtin.group import project_add_group
from zrb.task.decorator import python_task
from zrb.task.task import Task
from zrb.task.cmd_task import CmdTask
from zrb.task.resource_maker import ResourceMaker
from zrb.task_input.int_input import IntInput
from zrb.task_input.str_input import StrInput
from zrb.task_input.bool_input import BoolInput
from zrb.helper.accessories.name import get_random_name
from zrb.runner import runner

import os

CURRENT_DIR = os.path.dirname(__file__)

###############################################################################
# Task Input Definitions
###############################################################################

template_name_input = StrInput(
    name='template-name',
    shortcut='t',
    description='Template name',
    prompt='Template name',
    default=get_random_name()
)

base_image_input = StrInput(
    name='base-image',
    shortcut='i',
    description='Base image',
    prompt='Base image',
    default='stalchmst/moku:1.0.0'
)

default_app_port_input = IntInput(
    name='default-app-port',
    shortcut='p',
    description='Default app port',
    prompt='Default app port',
    default='8080'
)

build_custom_image_input = BoolInput(
    name='build-custom-image',
    description='Whether build custom image or not',
    prompt='Does user need to create custom image?',
    default=True
)

is_container_only_input = BoolInput(
    name='is-container-only',
    description='Whether app only run as container or not',
    prompt='Is this container only?',
    default=False
)

is_http_port_input = BoolInput(
    name='is-http-port',
    description='Whether app run on top of HTTP(s) protocol or not',
    prompt='Is this a web appp (run on top of HTTP(s))?',
    default=False
)

use_helm_input = BoolInput(
    name='use-helm',
    description='Whether using helm for deployment or not',
    prompt='Do you want to use helm for deployment?',
    default=False
)

###############################################################################
# Task Properties
###############################################################################

inputs = [
    project_dir_input,
    template_name_input,
    default_app_port_input,
    is_http_port_input,
    base_image_input,
    build_custom_image_input,
    is_container_only_input,
    use_helm_input,
]

replacements = {
    'zrbMetaTemplateName': '{{input.template_name}}',
    'zrbMetaBaseImage': '{{input.base_image}}',
    'zrbMetaDefaultAppPort': '{{input.default_app_port}}',
}

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='validate',
    inputs=inputs,
    retry=0
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_existing_project_dir(project_dir)
    template_name = kwargs.get('template_name')
    validate_inexisting_automation(project_dir, template_name)
    is_container_only_input: bool = kwargs.get('is_container_only')
    build_custom_image_input: bool = kwargs.get('build_custom_image')
    if not is_container_only_input and not build_custom_image_input:
        raise Exception(
            'Invalid options: Not is-container-only but not build-custom-image'
        )


copy_base_resource = ResourceMaker(
    name='copy-base-resource',
    inputs=inputs,
    upstreams=[validate],
    replacements=replacements,
    template_path=os.path.join(CURRENT_DIR, 'template', 'base'),
    destination_path='{{ input.project_dir }}',
    excludes=['*/__pycache__']
)

copy_http_port_resource = ResourceMaker(
    name='copy-http-port-resource',
    inputs=inputs,
    should_execute='{{ input.is_http_port }}',
    upstreams=[copy_base_resource],
    replacements=replacements,
    template_path=os.path.join(CURRENT_DIR, 'template', 'http-port'),
    destination_path='{{ input.project_dir }}',
    excludes=['*/__pycache__']
)

copy_custom_image_resource = ResourceMaker(
    name='copy-custom-image-resource',
    inputs=inputs,
    should_execute='{{ input.build_custom_image }}',
    upstreams=[copy_http_port_resource],
    replacements=replacements,
    template_path=os.path.join(CURRENT_DIR, 'template', 'build-custom-image'),
    destination_path='{{ input.project_dir }}',
    excludes=['*/__pycache__']
)

copy_http_port_custom_image_resource = ResourceMaker(
    name='copy-http-port-custom-image-resource',
    inputs=inputs,
    should_execute='{{ input.is_http_port and input.build_custom_image }}',
    upstreams=[copy_custom_image_resource],
    replacements=replacements,
    template_path=os.path.join(CURRENT_DIR, 'template', 'http-port-build-custom-image'),  # noqa
    destination_path='{{ input.project_dir }}',
    excludes=['*/__pycache__']
)

copy_helm_resource = ResourceMaker(
    name='copy-helm-resource',
    inputs=inputs,
    should_execute='{{ input.use_helm }}',
    upstreams=[copy_http_port_custom_image_resource],
    replacements=replacements,
    template_path=os.path.join(CURRENT_DIR, 'template', 'use-helm'),
    destination_path='{{ input.project_dir }}',
    excludes=['*/__pycache__']
)

copy_resource = CmdTask(
    name='copy-resource',
    upstreams=[
        copy_helm_resource,
    ],
    cmd='echo Resource copied',
)

register_module = create_register_module(
    module_path='_automate.generate_{{util.to_snake_case(input.template_name)}}.add',  # noqa
    alias='generate_{{util.to_snake_case(input.template_name)}}',
    inputs=[template_name_input],
    upstreams=[copy_resource]
)


@python_task(
    name='app-generator',
    group=project_add_group,
    upstreams=[register_module],
    inputs=inputs,
    runner=runner
)
async def add_app_generator(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')



from component.log import logger
from component.db_connection import engine

from module.auth.entity.group.repo import (
    GroupRepo, GroupDBRepo
)

group_repo: GroupRepo = GroupDBRepo(
    logger=logger, engine=engine
)



from zrb.task_input.str_input import StrInput
from zrb.task_input.int_input import IntInput
from zrb.helper.accessories.name import get_random_name
import os

SYSTEM_USER = os.getenv('USER', 'incognito')

project_dir_input = StrInput(
    name='project-dir',
    shortcut='d',
    description='Project directory',
    prompt='Project directory',
    default='.',
)

project_name_input = StrInput(
    name='project-name',
    shortcut='n',
    description='Project name',
    prompt='Project name (can be empty)',
    default='',
)

app_name_input = StrInput(
    name='app-name',
    shortcut='a',
    description='App name',
    prompt='App name',
    default='app',
)

app_image_default_namespace = os.getenv(
    'PROJECT_IMAGE_DEFAULT_NAMESPACE',
    'docker.io/' + SYSTEM_USER
)
app_image_name_input = StrInput(
    name='app-image-name',
    description='App image name',
    prompt='App image name',
    default=app_image_default_namespace + '/' + '{{util.to_kebab_case(input.app_name)}}',  # noqa
)

module_name_input = StrInput(
    name='module-name',
    shortcut='m',
    description='module name',
    prompt='module name',
    default='library',
)

entity_name_input = StrInput(
    name='entity-name',
    shortcut='e',
    description='Singular entity name',
    prompt='Singular entity name',
    default='book'
)

plural_entity_name_input = StrInput(
    name='plural-entity-name',
    description='Plural entity name',
    prompt='Plural entity name',
    default='books'
)

main_column_name_input = StrInput(
    name='column-name',
    shortcut='c',
    description='Main column name',
    prompt='Main column name',
    default='code'
)

column_name_input = StrInput(
    name='column-name',
    shortcut='c',
    description='Column name',
    prompt='Column name',
    default='title'
)

column_type_input = StrInput(
    name='column-type',
    shortcut='t',
    description='Column type',
    prompt='Column type',
    default='str'
)

task_name_input = StrInput(
    name='task-name',
    shortcut='t',
    description='Task name',
    prompt='Task name',
    default=f'run-{get_random_name()}',
)

http_port_input = IntInput(
    name='http-port',
    shortcut='p',
    description='HTTP port',
    prompt='HTTP port',
    default=8080,
)

env_prefix_input = StrInput(
    name='env-prefix',
    description='OS environment prefix',
    prompt='OS environment prefix',
    default='{{util.to_snake_case(util.coalesce(input.app_name, input.task_name, "MY")).upper()}}',  # noqa
)

package_name_input = StrInput(
    name='package-name',
    shortcut='p',
    description='Package name',
    prompt='Package name',
    default=get_random_name()
)

package_description_input = StrInput(
    name='package-description',
    description='Package description',
    prompt='Package description',
    default='My super package'
)

package_homepage_input = StrInput(
    name='package-homepage',
    description='Package homepage',
    prompt='Package homepage',
    default='https://github.com/' + SYSTEM_USER + '/{{util.to_kebab_case(input.package_name)}}'  # noqa
)

package_bug_tracker_input = StrInput(
    name='package-bug-tracker',
    description='Package bug tracker',
    prompt='Package bug tracker',
    default='https://github.com/' + SYSTEM_USER + '/{{util.to_kebab_case(input.package_name)}}/issues' # noqa
)

package_author_name_input = StrInput(
    name='package-author-name',
    prompt='Package author name',
    description='Package author name',
    default=SYSTEM_USER
)

package_author_email_input = StrInput(
    name='package-author-email',
    prompt='Package author email',
    description='Package author email',
    default=f'{SYSTEM_USER}@gmail.com'
)



from zrb import CmdTask, runner
from zrb.builtin.group import project_group
from .image import push_snake_zrb_app_name_image
from ._config import CURRENT_DIR, DEPLOYMENT_DIR
from ._input import (
    enable_monitoring_input, image_input, deploy_mode_input, pulumi_stack_input
)
from ._env import (
    image_env, pulumi_backend_url_env, pulumi_config_passphrase_env,
    deployment_modules_env, deployment_mode_env,
    deployment_enable_monitoring_env
)
from ._env_file import deployment_app_env_file, deployment_config_env_file
import os

###############################################################################
# ⚙️ deploy-kebab-zrb-task-name
###############################################################################

deploy_snake_zrb_app_name = CmdTask(
    icon='🚧',
    name='deploy-kebab-zrb-app-name',
    description='Deploy human readable zrb app name',
    group=project_group,
    inputs=[
        image_input,
        pulumi_stack_input,
        deploy_mode_input,
        enable_monitoring_input,
    ],
    upstreams=[push_snake_zrb_app_name_image],
    cwd=DEPLOYMENT_DIR,
    env_files=[
        deployment_config_env_file,
        deployment_app_env_file,
    ],
    envs=[
        pulumi_backend_url_env,
        pulumi_config_passphrase_env,
        image_env,
        deployment_modules_env,
        deployment_mode_env,
        deployment_enable_monitoring_env,
    ],
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'pulumi-init-stack.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'pulumi-up.sh'),
    ]
)
runner.register(deploy_snake_zrb_app_name)

###############################################################################
# ⚙️ destroy-kebab-zrb-task-name
###############################################################################

destroy_snake_zrb_app_name = CmdTask(
    icon='💨',
    name='destroy-kebab-zrb-app-name',
    description='Remove human readable zrb app name deployment',
    group=project_group,
    inputs=[
        pulumi_stack_input,
    ],
    cwd=DEPLOYMENT_DIR,
    env_files=[
        deployment_config_env_file,
        deployment_app_env_file,
    ],
    envs=[
        pulumi_backend_url_env,
        pulumi_config_passphrase_env,
        image_env,
        deployment_modules_env,
    ],
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'pulumi-init-stack.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'pulumi-destroy.sh'),
    ]
)
runner.register(destroy_snake_zrb_app_name)



from zrb import CmdTask, Env, EnvFile, IntInput, StrInput, runner
from zrb.builtin.group import project_group
from .image import push_snake_zrb_app_name_image, image_input, image_env
from ._common import (
    CURRENT_DIR, DEPLOYMENT_DIR, APP_TEMPLATE_ENV_FILE_NAME,
    DEPLOYMENT_TEMPLATE_ENV_FILE_NAME
)
import os

###############################################################################
# 🔤 Input Definitions
###############################################################################

replica_input = IntInput(
    name='kebab-zrb-app-name-replica',
    description='Replica of "kebab-zrb-app-name"',
    prompt='Replica of "kebab-zrb-app-name"',
    default=1,
)

pulumi_stack_input = StrInput(
    name='kebab-zrb-app-name-pulumi-stack',
    description='Pulumi stack name for "kebab-zrb-app-name"',
    prompt='Pulumi stack name for "kebab-zrb-app-name"',
    default=os.getenv('ZRB_ENV', 'dev')
)

###############################################################################
# 🌳 Env File Definitions
###############################################################################

deployment_app_env_file = EnvFile(
    path=APP_TEMPLATE_ENV_FILE_NAME,
    prefix='DEPLOYMENT_APP_ZRB_ENV_PREFIX'
)

deployment_config_env_file = EnvFile(
    path=DEPLOYMENT_TEMPLATE_ENV_FILE_NAME,
    prefix='DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX'
)

###############################################################################
# 🌱 Env Definitions
###############################################################################

pulumi_backend_url_env = Env(
    name='PULUMI_BACKEND_URL',
    os_name='PULUMI_ZRB_ENV_PREFIX_BACKEND_URL',
    default=f'file://{DEPLOYMENT_DIR}/state'
)

pulumi_config_passphrase_env = Env(
    name='PULUMI_CONFIG_PASSPHRASE',
    os_name='PULUMI_ZRB_ENV_PREFIX_CONFIG_PASSPHRASE',
    default='secret'
)

deployment_replica_env = Env(
    name='REPLICA',
    os_name='DEPLOYMENT_ZRB_ENV_PREFIX',
    default='{{input.snake_zrb_app_name_replica}}'
)

###############################################################################
# ⚙️ deploy-kebab-zrb-task-name
###############################################################################

deploy_snake_zrb_app_name = CmdTask(
    icon='🚧',
    name='deploy-kebab-zrb-app-name',
    description='Deploy human readable zrb app name',
    group=project_group,
    inputs=[
        image_input,
        replica_input,
        pulumi_stack_input,
    ],
    upstreams=[push_snake_zrb_app_name_image],
    cwd=DEPLOYMENT_DIR,
    env_files=[
        deployment_config_env_file,
        deployment_app_env_file,
    ],
    envs=[
        pulumi_backend_url_env,
        pulumi_config_passphrase_env,
        image_env,
        deployment_replica_env,
    ],
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'pulumi-init-stack.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'pulumi-up.sh'),
    ]
)
runner.register(deploy_snake_zrb_app_name)

###############################################################################
# ⚙️ destroy-kebab-zrb-task-name
###############################################################################

destroy_snake_zrb_app_name = CmdTask(
    icon='💨',
    name='destroy-kebab-zrb-app-name',
    description='Remove human readable zrb app name deployment',
    group=project_group,
    inputs=[
        pulumi_stack_input,
    ],
    cwd=DEPLOYMENT_DIR,
    env_files=[
        deployment_config_env_file,
        deployment_app_env_file,
    ],
    envs=[
        pulumi_backend_url_env,
        pulumi_config_passphrase_env,
        image_env,
        deployment_replica_env,
    ],
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'pulumi-init-stack.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'pulumi-destroy.sh'),
    ]
)
runner.register(destroy_snake_zrb_app_name)



from zrb.task.task import Task
from zrb.task.cmd_task import CmdTask
from zrb.task_input.str_input import StrInput


def test_task_input():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        favorite_drink = kwargs['favorite_drink']
        return f'hello {name}, your favorite drink is {favorite_drink}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name'),
            StrInput(name='favorite-drink')
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function(name='Dumbledore', favorite_drink='Elixir')
    assert result == 'hello Dumbledore, your favorite drink is Elixir'


def test_cmd_task_input():
    task = CmdTask(
        name='hello-name',
        inputs=[
            StrInput(name='name'),
            StrInput(name='favorite-drink')
        ],
        cmd='echo hello $_INPUT_NAME, your favorite drink is $_INPUT_FAVORITE_DRINK',  # noqa
        retry=0
    )
    function = task.to_function()
    result = function(name='Dumbledore', favorite_drink='Elixir')
    assert result.output == 'hello Dumbledore, your favorite drink is Elixir'


def test_task_input_with_default_value():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        favorite_drink = kwargs['favorite_drink']
        return f'hello {name}, your favorite drink is {favorite_drink}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(name='favorite-drink', default='Elixir')
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result == 'hello Nicholas Flamel, your favorite drink is Elixir'


def test_cmd_task_input_with_default_value():
    task = CmdTask(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(name='favorite-drink', default='Elixir')
        ],
        cmd='echo hello $_INPUT_NAME, your favorite drink is $_INPUT_FAVORITE_DRINK',  # noqa
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result.output == 'hello Nicholas Flamel, your favorite drink is Elixir'  # noqa


def test_task_input_with_jinja_value():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        alias = kwargs['alias']
        return f'hello {name}, aka {alias}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(name='alias', default='{{input.name}}')
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result == 'hello Nicholas Flamel, aka Nicholas Flamel'


def test_cmd_task_input_with_jinja_value():
    task = CmdTask(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(name='alias', default='{{input.name}}')
        ],
        cmd='echo hello $_INPUT_NAME, aka $_INPUT_ALIAS',
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result.output == 'hello Nicholas Flamel, aka Nicholas Flamel'


def test_task_input_with_should_not_be_rendered_jinja_value():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        alias = kwargs['alias']
        return f'hello {name}, aka {alias}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(
                name='alias', default='{{input.name}}', should_render=False
            )
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result == 'hello Nicholas Flamel, aka {{input.name}}'


def test_cmd_task_input_with_should_not_be_rendered_jinja_value():
    task = CmdTask(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(
                name='alias', default='{{input.name}}', should_render=False
            )
        ],
        cmd='echo hello $_INPUT_NAME, aka $_INPUT_ALIAS',
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result.output == 'hello Nicholas Flamel, aka {{input.name}}'


def test_task_input_with_jinja_value_and_partial_custom_kwargs():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        alias = kwargs['alias']
        return f'hello {name}, aka {alias}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(name='alias', default='{{input.name}}')
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function(name='Alchemist')
    assert result == 'hello Alchemist, aka Alchemist'


def test_task_input_with_jinja_value_and_custom_kwargs():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        alias = kwargs['alias']
        return f'hello {name}, aka {alias}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(name='alias', default='{{input.name}}')
        ],
        run=_run,
        retry=0
    )
    # Name and alias provided
    function = task.to_function()
    result = function(name='Nicholas Flamel', alias='Alchemist')
    assert result == 'hello Nicholas Flamel, aka Alchemist'


def test_task_input_with_the_same_name():
    task = Task(
        name='task',
        inputs=[
            StrInput(name='name', default='articuno'),
            StrInput(name='name', default='zapdos'),
            StrInput(name='name', default='moltres'),
        ],
        run=lambda *args, **kwargs: kwargs.get('name')
    )
    function = task.to_function()
    result = function()
    assert result == 'moltres'


def test_task_input_with_the_same_name_on_upstream():
    task_upstream = Task(
        name='task-upstream',
        inputs=[
            StrInput(name='name', default='articuno')
        ],
        run=lambda *args, **kwargs: kwargs.get('name')
    )
    task = Task(
        name='task',
        inputs=[
            StrInput(name='name', default='zapdos')
        ],
        upstreams=[task_upstream],
        run=lambda *args, **kwargs: kwargs.get('name')
    )
    function = task.to_function()
    result = function()
    assert result == 'zapdos'




from typing import Any, List
from zrb import Task

###############################################################################
# Functions
###############################################################################


def should_start_local_microservices(*args: Any, **kwargs: Any) -> bool:
    if not kwargs.get('local_snake_zrb_app_name', True):
        return False
    run_mode = kwargs.get('snake_zrb_app_name_run_mode', 'monolith')
    return run_mode == 'microservices'


def get_support_container_compose_profiles(
    *args: Any, **kwargs: Any
) -> List[str]:
    task: Task = kwargs.get('_task')
    env_map = task.get_env_map()
    compose_profiles: List[str] = []
    broker_type = env_map.get('APP_BROKER_TYPE', 'rabbitmq')
    if broker_type in ['rabbitmq', 'kafka']:
        compose_profiles.append(broker_type)
    if kwargs.get('enable_snake_zrb_app_name_monitoring', False):
        compose_profiles.append('monitoring')
    return compose_profiles


def get_container_compose_profiles(*args: Any, **kwargs: Any) -> str:
    compose_profiles = get_support_container_compose_profiles(*args, **kwargs)
    compose_profiles.append(
        kwargs.get('snake_zrb_app_name_run_mode', 'monolith'),
    )
    return compose_profiles


def activate_support_compose_profile(*args: Any, **kwargs: Any) -> str:
    compose_profiles = get_support_container_compose_profiles(*args, **kwargs)
    compose_profile_str = ','.join(compose_profiles)
    return f'export COMPOSE_PROFILES={compose_profile_str}'


def should_start_support_container(*args: Any, **kwargs: Any) -> bool:
    if not kwargs.get('local_snake_zrb_app_name', True):
        return False
    compose_profiles = get_support_container_compose_profiles(*args, **kwargs)
    return len(compose_profiles) > 0


def should_start_local_monolith(*args: Any, **kwargs: Any) -> bool:
    if not kwargs.get('local_snake_zrb_app_name', True):
        return False
    return kwargs.get('snake_zrb_app_name_run_mode', 'monolith') == 'monolith'


def activate_selected_compose_profile(*args: Any, **kwargs: Any) -> str:
    compose_profiles = get_container_compose_profiles(*args, **kwargs)
    compose_profile_str = ','.join(compose_profiles)
    return f'export COMPOSE_PROFILES={compose_profile_str}'


def activate_all_compose_profile(*args: Any, **kwargs: Any) -> str:
    compose_profile_str = ','.join([
        'monitoring', 'monolith', 'microservices', 'kafka', 'rabbitmq'
    ])
    return f'export COMPOSE_PROFILES={compose_profile_str}'


def should_start_container(*args: Any, **kwargs: Any) -> bool:
    return kwargs.get('local_snake_zrb_app_name', True)



from zrb.task.task import Task
from zrb.task.parallel import Parallel


def test_task_shift_right_operator():
    task_1 = Task(name='task-1')
    task_2 = Task(name='task-2')
    task_3 = Task(name='task-3')
    task_4 = Task(name='task-4')
    # define DAG
    task_1 >> task_2 >> task_3 >> task_4
    # test 1
    upstream_1 = task_1._get_upstreams()
    assert len(upstream_1) == 0
    # test 2
    upstream_2 = task_2._get_upstreams()
    assert len(upstream_2) == 1
    assert upstream_2[0] == task_1
    # test 3
    upstream_3 = task_3._get_upstreams()
    assert len(upstream_3) == 1
    assert upstream_3[0] == task_2
    # test 4
    upstream_4 = task_4._get_upstreams()
    assert len(upstream_4) == 1
    assert upstream_4[0] == task_3


def test_task_shift_right_operator_with_parallel():
    task_1 = Task(name='task-1')
    task_2 = Task(name='task-2')
    task_3 = Task(name='task-3')
    task_4 = Task(name='task-4')
    task_5 = Task(name='task-5')
    task_6 = Task(name='task-6')
    # define DAG
    task_1 >> Parallel(task_2, task_3) >> Parallel(task_4, task_5) >> task_6
    # test 1
    upstream_1 = task_1._get_upstreams()
    assert len(upstream_1) == 0
    # test 2
    upstream_2 = task_2._get_upstreams()
    assert len(upstream_2) == 1
    assert upstream_2[0] == task_1
    # test 3
    upstream_3 = task_3._get_upstreams()
    assert len(upstream_3) == 1
    assert upstream_3[0] == task_1
    # test 4
    upstream_4 = task_4._get_upstreams()
    assert len(upstream_4) == 2
    assert upstream_4[0] == task_2
    assert upstream_4[1] == task_3
    # test 5
    upstream_5 = task_5._get_upstreams()
    assert len(upstream_5) == 2
    assert upstream_5[0] == task_2
    assert upstream_5[1] == task_3
    # test 6
    upstream_6 = task_6._get_upstreams()
    assert len(upstream_6) == 2
    assert upstream_6[0] == task_4
    assert upstream_6[1] == task_5



from zrb import (
    DockerComposeTask, Env, EnvFile, ServiceConfig, runner, PortChecker,
)
from zrb.builtin.group import project_group
from ._common import (
    RESOURCE_DIR, APP_DIR, local_input, host_input
)
from .image import image_input, image_env, build_snake_zrb_app_name_image
import os

###############################################################################
# 🌳 Env File Definitions
###############################################################################

compose_env_file = EnvFile(
    path=os.path.join(RESOURCE_DIR, 'docker-compose.env'),
    prefix='CONTAINER_ZRB_ENV_PREFIX'
)

###############################################################################
# 🌱 Env Definitions
###############################################################################

host_port_env = Env(
    name='HOST_PORT',
    os_name='CONTAINER_ZRB_ENV_PREFIX_HOST_PORT',
    default='zrbAppPort'
)

###############################################################################
# Service Config Definitions
###############################################################################

snake_zrb_app_name_service_config = ServiceConfig(
    env_files=[
        EnvFile(
            path=os.path.join(APP_DIR, 'template.env'),
            prefix='CONTAINER_ZRB_ENV_PREFIX'
        )
    ]
)

###############################################################################
# ⚙️ remove-zrb-task-name-container
###############################################################################

remove_snake_zrb_app_name_container = DockerComposeTask(
    icon='💨',
    name='remove-kebab-zrb-app-name-container',
    description='Remove human readable zrb app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    compose_cmd='down',
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs={
        'kebab-zrb-app-name': snake_zrb_app_name_service_config
    },
    env_files=[compose_env_file],
    envs=[
        image_env,
        host_port_env,
    ],
)
runner.register(remove_snake_zrb_app_name_container)

###############################################################################
# ⚙️ stop-zrb-task-name-container
###############################################################################

stop_snake_zrb_app_name_container = DockerComposeTask(
    icon='⛔',
    name='stop-kebab-zrb-app-name-container',
    description='Stop human readable zrb app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    compose_cmd='stop',
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs={
        'kebab-zrb-app-name': snake_zrb_app_name_service_config
    },
    env_files=[compose_env_file],
    envs=[
        image_env,
        host_port_env,
    ],
)
runner.register(stop_snake_zrb_app_name_container)

###############################################################################
# ⚙️ init-zrb-task-name-container
###############################################################################

init_snake_zrb_app_name_container = DockerComposeTask(
    icon='🔥',
    name='init-kebab-zrb-app-name-container',
    inputs=[
        local_input,
        host_input,
        image_input,
    ],
    should_execute='{{input.local_snake_zrb_app_name}}',
    upstreams=[
        remove_snake_zrb_app_name_container,
        build_snake_zrb_app_name_image,
    ],
    cwd=RESOURCE_DIR,
    compose_cmd='up',
    compose_flags=['-d'],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs={
        'kebab-zrb-app-name': snake_zrb_app_name_service_config
    },
    env_files=[compose_env_file],
    envs=[
        image_env,
        host_port_env,
    ],
)

###############################################################################
# ⚙️ start-zrb-task-name-container
###############################################################################

start_snake_zrb_app_name_container = DockerComposeTask(
    icon='🐳',
    name='start-kebab-zrb-app-name-container',
    description='Start human readable zrb app name container',
    group=project_group,
    inputs=[
        local_input,
        host_input,
        image_input,
    ],
    should_execute='{{input.local_snake_zrb_app_name}}',
    upstreams=[init_snake_zrb_app_name_container],
    cwd=RESOURCE_DIR,
    compose_cmd='logs',
    compose_flags=['-f'],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs={
        'kebab-zrb-app-name': snake_zrb_app_name_service_config
    },
    env_files=[compose_env_file],
    envs=[
        image_env,
        host_port_env,
    ],
    checkers=[
        PortChecker(
            name='check-kebab-zrb-app-name',
            host='{{input.snake_zrb_app_name_host}}',
            port='{{env.HOST_PORT}}',
        ),
    ]
)
runner.register(start_snake_zrb_app_name_container)



from zrb.helper.typing import Any, List, Mapping
from abc import ABC, abstractmethod

# flake8: noqa E501

class AnyInput(ABC):
    '''
    Abstract base class representing a generalized input specification.
    This class serves as a template for creating various input types, 
    providing a standardized interface for input handling and processing.

    '''

    @abstractmethod
    def get_name(self) -> str:
        '''
        Retrieves the name of the input.

        Returns:
            str: The name of the input.
        '''
        pass

    @abstractmethod
    def get_default(self) -> Any:
        '''
        Obtains the default value of the input.

        Returns:
            Any: The default value of the input. The type can be any, depending on the input specification.
        '''
        pass

    @abstractmethod
    def get_param_decl(self) -> List[str]:
        '''
        Fetches a list of parameter option associated with the input (i.e., `-f` or `--file`).

        Returns:
            List[str]: A list containing strings of parameter options.
        '''
        pass

    @abstractmethod
    def get_options(self) -> Mapping[str, Any]:
        '''
        Provides a mapping (dictionary) representing the input.

        Returns:
            Mapping[str, Any]: A dictionary where keys are option names and values are the corresponding details.
        '''
        pass

    @abstractmethod
    def should_render(self) -> bool:
        '''
        Determines whether or not the input should be rendered.

        Returns:
            bool: True if the input should be rendered, False otherwise.
        '''
        pass

    @abstractmethod
    def is_hidden(self) -> bool:
        '''
        Checks whether the input value is meant to be hidden from view or output.

        Returns:
            bool: True if the input is hidden, False otherwise.
        '''
        pass



from zrb.builtin.md5 import hash_text, sum_file
import pathlib
import os


def test_sum_file():
    dir_path = pathlib.Path(__file__).parent.absolute()
    function = sum_file.to_function()
    result = function(file=os.path.join(dir_path, 'md5_test_sum.txt'))
    assert result == '5f935c38df842258336f683502ac153d'


def test_hash_text():
    function = hash_text.to_function()
    result = function(text='Philosopher Stone')
    assert result == '5f935c38df842258336f683502ac153d'



from zrb.helper.typing import Any
from zrb.builtin.group import md5_group
from zrb.task.decorator import python_task
from zrb.task_input.str_input import StrInput
from zrb.runner import runner

import aiofiles
import hashlib

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='hash',
    group=md5_group,
    inputs=[
        StrInput(
            name='text',
            shortcut='t',
            description='Text',
            default=''
        )
    ],
    description='Hash md5',
    retry=0,
    runner=runner
)
async def hash_text(*args: str, **kwargs: Any):
    text: str = kwargs.get('text', '')
    hashed_text = hashlib.md5(text.encode()).hexdigest()
    return hashed_text


@python_task(
    name='sum',
    group=md5_group,
    inputs=[
        StrInput(
            name='file',
            shortcut='f',
            description='File',
            default=''
        )
    ],
    description='Sum md5 file',
    retry=0,
    runner=runner
)
async def sum_file(*args: str, **kwargs: Any):
    file_path: str = kwargs.get('file', '')
    async with aiofiles.open(file_path, mode='rb') as file:
        contents = await file.read()
    hashed_text = hashlib.md5(contents).hexdigest()
    return hashed_text



from zrb.config.config import logging_level
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Mapping, Optional
from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from functools import lru_cache

import logging
import os

_PROJECT_DIR_MAP: Mapping[str, str] = {}


def inject_default_env():
    # Inject PYTHONUNBUFFERED
    if 'PYTHONUNBUFFERED' not in os.environ:
        if logging_level <= logging.INFO:
            logger.info(colored('Set PYTHONUNBUFFERED to 1', attrs=['dark']))
        os.environ['PYTHONUNBUFFERED'] = '1'
    # Inject ZRB_HOME_DIR
    if 'ZRB_HOME_DIR' not in os.environ:
        default_home_dir = os.path.dirname(os.path.dirname(__file__))
        if logging_level <= logging.INFO:
            logger.info(colored(
                f'Set ZRB_HOME_DIR to {default_home_dir}', attrs=['dark']
            ))
        os.environ['ZRB_HOME_DIR'] = default_home_dir
    # Inject ZRB_PROJECT_DIR
    current_dir = os.getcwd()
    if current_dir not in _PROJECT_DIR_MAP:
        if logging_level <= logging.INFO:
            logger.info(colored('Getting project directory', attrs=['dark']))
        zrb_project_dir = _get_project_dir(current_dir)
        if zrb_project_dir is None:
            zrb_project_dir = current_dir
        _PROJECT_DIR_MAP[current_dir] = zrb_project_dir
        if logging_level <= logging.INFO:
            logger.info(colored(
                f'Set ZRB_PROJECT_DIR to {zrb_project_dir}', attrs=['dark']
            ))
        os.environ['ZRB_PROJECT_DIR'] = zrb_project_dir
    # Inject ZRB_PROJECT_NAME
    if 'ZRB_PROJECT_NAME' not in os.environ:
        zrb_project_name = os.path.basename(zrb_project_dir)
        if logging_level <= logging.INFO:
            logger.info(colored(
                f'Set ZRB_PROJECT_NAME to {zrb_project_name}', attrs=['dark']
            ))
        os.environ['ZRB_PROJECT_NAME'] = zrb_project_name


@lru_cache
@typechecked
def _get_project_dir(project_dir: str) -> Optional[str]:
    project_script = os.path.join(project_dir, 'zrb_init.py')
    if os.path.isfile(project_script):
        return project_dir
    # zrb_init.py not found, look for it on the parent directory
    parent_project_dir = os.path.dirname(project_dir)
    if parent_project_dir == project_dir:
        return None
    return _get_project_dir(parent_project_dir)
